import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)  

import numpy as np
import scipy as sp

import matplotlib as mplt
import matplotlib.pyplot as plt

import itertools
from scipy.spatial import cKDTree



def get_pairs_d(pos, d0=None,
                d_eps=None, d_min=None, d_max=None,
                bins=None, hist_range=None, redf=None, doplot=False, ax=None):
    """
    From a list of positions, finds the most common distance between them (d0),
    and pairs the points that are at such distance. If d0 is given, it is used instead of computing it.
    
    The pairs are ordered such that for pair (p1, p2), p1 is always to the left (smaller x value) than p2.
    """

    d_eps = d_eps or 0.8
    d_min = d_min or 0
    d_max = d_max or 60

    if bins is None:
        if redf is not None:
            bins = int( 0.75 * max(redf.data.shape) )
        else:
            raise ValueError("bins must be specified if redf is not given")

    if hist_range is None:
        if redf is not None:
            hist_range = (0, min(redf.data.shape))
        else:
            raise ValueError("hist_range must be specified if redf is not given")
        
    if pos is None or len(pos) < 2:
        return [], [], None, None

    pairs = list(itertools.combinations(pos, 2))
    distances = [np.linalg.norm(p1-p2) for p1,p2 in pairs]

    if d0 is None:
        hist, edges = np.histogram(distances, bins=bins, range=hist_range)
        centers = (edges[:-1]+edges[1:])/2

        idx = (d_min <= centers) & (centers <= d_max)
        idx_max = np.argmax(hist[idx])
        d0 = centers[idx][idx_max]

    paired = [(p1,p2) for p1,p2 in pairs if np.abs(np.linalg.norm(p1-p2)-d0) < d_eps]
    paired = [[p1,p2]  if p1[0]>p2[0] else [p2,p1] for (p1,p2) in paired]

    if len(paired) == 0:
        return [], [], d0, None

    list1, list2 = list(zip(*paired))
    
    pos1 = np.array(list1)
    pos2 = np.array(list2)
    disp_sign = np.mean(pos2-pos1, axis=0)

    # Plotting (optional)
    if doplot:
        if ax is None:
            ax = plt.gca()
        cnts, edges, bars = ax.hist(distances,  bins=bins, range=hist_range)
        ax.axvline(x=d0, color='r', linestyle='--', linewidth=1, alpha=0.5)
        bars[np.argmax(cnts)].set_facecolor('red')

    return list1, list2, d0, disp_sign



def get_pairs_dxy(pos, disp=None, 
                  dx_eps=None, dy_eps=None, d_eps=None, dx_min=None, dx_max=None, dy_min=None, dy_max=None, d_min=None,
                  bins=None, hist_range=None, redf=None, doplot=False, axs=None, fig=None):
    """
    From a list of positions, finds the most common distances between them in both x and y axes (disp),
    and pairs the points that are at such distances.

    If disp is given, it is used as the most common distance in both axes instead of computing it.
    
    The pairs are ordered such that for pair (p1, p2), p1 is always to the left (smaller x value) than p2.
    
    Note: this function is similar to get_pairs_d(), but finds the most common distances both in x and y axes.
    """
    
    dx_eps = dx_eps or 0.8
    dy_eps = dy_eps or 0.8
    d_eps = d_eps or 0.8
    dx_min = dx_min or 0
    dx_max = dx_max or 60
    dy_min = dy_min or 0
    dy_max = dy_max or 60
    d_min = d_min or 0

        
    if pos is None or len(pos) < 2:
        return [], [], None, None
    
    pairs = list(itertools.combinations(pos, 2))

    if disp is None:

        if bins is None:
            if redf is not None:
                bins = int( 0.75 * max(redf.data.shape) )
            else:
                raise ValueError("bins must be specified if redf is not given")

        if hist_range is None:
            if redf is not None:
                hist_range = (0, min(redf.data.shape))
            else:
                raise ValueError("hist_range must be specified if redf is not given")
        
        disp = list()
        for i, d_min, d_max in zip([0, 1], [dx_min, dy_min], [dx_max, dy_max]): # for each axis
            distances = [abs(p1[i]-p2[i]) for p1,p2 in pairs]

            hist, edges = np.histogram(distances, bins=bins, range=hist_range)
            centers = (edges[:-1]+edges[1:])/2

            idx = (d_min <= centers) & (centers <= d_max)
            idx_max = np.argmax(hist[idx])
            d0 = centers[idx][idx_max]

            disp.append(d0)

    paired = [(p1,p2) for p1,p2 in pairs if ( abs( abs( p1[0] - p2[0] ) - disp[0] ) < dx_eps and abs( (abs( p1[1] - p2[1] ) - disp[1] ) ) < dy_eps )]
    paired = [[p1,p2]  if p1[0]>p2[0] else [p2,p1] for (p1,p2) in paired]

    if len(paired) == 0:
        return [], [], disp, None
    
    list1, list2 = list(zip(*paired))
    
    pos1 = np.array(list1)
    pos2 = np.array(list2)
    disp_sign = np.mean(pos2-pos1, axis=0)

    # Plotting (optional)
    if doplot:
        if axs is None:
            if fig is None:
                fig = plt.gcf()
            axs = fig.subplots(nrows=2, sharex=True)

        if len(axs) == 1:
            ax = axs[0]
            for i, color in zip([0, 1], ['r','b']):
                distances = [abs(p1[i]-p2[i]) for p1,p2 in pairs]
                cnts, edges, bars = ax.hist(distances, bins=bins, range=hist_range, alpha=0.3)
                ax.axvline(x=disp[i], color=color, linestyle='--', linewidth=1, alpha=0.3)
                bars[np.argmax(cnts)].set_facecolor('red')
        elif len(axs) == 2:
            for i in [0, 1]:
                distances = [abs(p1[i]-p2[i]) for p1,p2 in pairs]
                cnts, edges, bars = axs[i].hist(distances, bins=bins, range=hist_range)
                axs[i].axvline(x=disp[i], color='k', linestyle='--', linewidth=1, alpha=0.5)
                bars[np.argmax(cnts)].set_facecolor('red')
        else:
            raise ValueError("axs must be a list of length 1 or 2")

    return list1, list2, disp, disp_sign



def get_best_pairs(list1, list2, disp_sign, dist_err=None, disp_sign_err=None):
    """
    From two lists which correspond to paired points, if there are points participating in more
    than one pair, return only the best pair according to displacement disp_sign. 
    
    If dist_err (scalar) or disp_sign_err are given, only pairs whose points are displace by disp_sign within 
    a distance of dist_err (1d) or within disp_sign_err (2d) are considered.

    Example
    -------
    An example of pair finding where first we detect the sources and find the pairs without a priori knowledge of the displacement.
    ```
    # Detect sources in the sky (pos_seg is a list of (x,y) positions)

    # Define some parameters
    
    d_eps = 0.8
    bins = int( 0.75 * max(redf.data.shape) )
    hist_range = (0, min(redf.data.shape))

    ## Find pairs of sources matching in scalar distance, using the most common distance between sources
    seg1, seg2, seg_d0, seg_disp_sign = get_pairs_d(pos_seg, d_eps=d_eps, bins=bins, hist_range=hist_range)

    ## Get only the best pairs, according to the displacement sign
    seg1_best, seg2_best, seg_disp_best, seg_disp_sign_best = get_best_pairs(seg1, seg2, seg_disp_sign)

    ## Find pairs of sources matching in 2d position, using the most common displacement between sources
    seg1xy, seg2xy, seg_disp_xy, seg_disp_sign_xy = get_pairs_dxy(pos_seg, d_eps=d_eps, bins=bins, hist_range=hist_range)

    ## Get only the best pairs, according to the displacement sign
    seg1xy_best, seg2xy_best, seg_disp_xy_best, seg_disp_sign_xy_best = get_best_pairs(seg1xy, seg2xy, seg_disp_sign_xy)
    ```
    Alternatively, using a priori knowledge of the displacement, we can find the pairs directly:
    ```
    # Get the average displacement between pairs and its std from already calibrated sources
    disp_mean = np.mean([redf.astrometry_info[-1]['seg_disp_sign_xy'] for redf in ReducedFit.objects.filter(flags__has=ReducedFit.FLAGS.BUILT_REDUCED, obsmode="POLARIMETRY") if 'seg_disp_sign_xy' in redf.astrometry_info[-1] and isinstance(redf.astrometry_info[-1]['seg_disp_sign_xy'], np.ndarray)], axis=0)
    disp_std = np.std([redf.astrometry_info[-1]['seg_disp_sign_xy'] for redf in ReducedFit.objects.filter(flags__has=ReducedFit.FLAGS.BUILT_REDUCED, obsmode="POLARIMETRY") if 'seg_disp_sign_xy' in redf.astrometry_info[-1] and isinstance(redf.astrometry_info[-1]['seg_disp_sign_xy'], np.ndarray)], axis=0)

    # Detect sources (pos_seg is a list of (x,y) positions)

    # Directly find the best pairs with the mean disp_mmmean and disp_std
    seg1xy_best, seg2xy_best, seg_disp_xy_best, seg_disp_sign_xy_best = get_best_pairs(pos_seg, pos_seg, disp_mean, disp_sign_err=5*disp_std)
    ```

    
    Parameters
    ----------
    list1, list2 : list, list
        paired lists of points
    disp_sign: (float, float)
        the displacement between points
    
    Returns
    -------
    list1, list2 : list, list
        paired list of points
    d0_new : float
        recomputed points
    disp_sign_new : 
        recomputed displacement
    """
    
    if list1 is None or len(list1) < 2:
        return [], [], None, None

    set1 = {tuple(p1) for p1 in list1}
    set2 = {tuple(p2) for p2 in list2}

    def get_best_companion(p,pL):
        p = np.array(p) ## so now p+disp_sign-x is an array even if disp_sign and x are tuples
        return min(pL, key=lambda x: np.abs(np.linalg.norm(p+disp_sign-x)))

    paired = [(p1, get_best_companion(p1,set2)) for p1 in set1]

    if dist_err is not None:
        paired = [pair for pair in paired if np.linalg.norm(pair[0]+disp_sign-pair[1]) < dist_err]
    
    if disp_sign_err is not None:
        paired = [pair for pair in paired if np.abs(pair[0][0]+disp_sign[0]-pair[1][0]) < disp_sign_err[0]]
        paired = [pair for pair in paired if np.abs(pair[0][1]+disp_sign[1]-pair[1][1]) < disp_sign_err[1]]

    if len(paired) == 0:
        return [], [], None, None

    list1, list2 = list(zip(*paired))
    
    pos1 = np.array(list1)
    pos2 = np.array(list2)
    disp_sign_new = np.mean(pos2-pos1, axis=0)
    d0_new = np.linalg.norm(disp_sign_new)
    
    return list1, list2, d0_new, disp_sign_new




import pytest
from pathlib import Path

from .conftest import TEST_CONFIG

# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_path=TEST_CONFIG)

# other imports
import os
from pytest import approx
import numpy as np

# logging
import logging
logger = logging.getLogger(__name__)

# fixtures
from .fixtures import load_test_catalog


@pytest.mark.django_db(transaction=True)
def test_polarimetry(load_test_catalog):

    from iop4lib.db import Epoch, RawFit, ReducedFit, AstroSource, PhotoPolResult
    from iop4lib.enums import IMGTYPES, INSTRUMENTS

    # get epochs that have DIPOL observations in them
    from iop4lib.iop4 import list_local_epochnames
    epochname_L = list_local_epochnames()
    epoch_L = [Epoch.create(epochname=epochname) for epochname in epochname_L]
    epoch_L = [epoch for epoch in epoch_L if epoch.rawfits.filter(instrument=INSTRUMENTS.DIPOL).exists()]

    for epoch in epoch_L:
        epoch.build_master_biases()

    for epoch in epoch_L:
        epoch.build_master_darks()

    for epoch in epoch_L:
        epoch.build_master_flats()
    
    rawfits = RawFit.objects.filter(epoch__in=epoch_L, instrument=INSTRUMENTS.DIPOL, imgtype=IMGTYPES.LIGHT).all()
    Epoch.reduce_rawfits(rawfits)

    for epoch in epoch_L:
        epoch.compute_relative_polarimetry()

    qs_res = PhotoPolResult.objects.filter(epoch__in=epoch_L).all()

    # HD 204827 (polarization standard)
    # https://www.not.iac.es/instruments/turpol/std/hpstd.html
    # pL(R): 4.893 +- 0.029 %, ChiL(R) 59.10 +- 0.17

    res = qs_res.filter(instrument=INSTRUMENTS.DIPOL, epoch__night="2023-10-25").get(astrosource__name="HD 204827")
    assert res.p == approx(4.893/100, abs=max(2*res.p_err, 0.5/100)) 
    assert res.chi == approx(59.10, abs=max(2*res.chi_err, 5))

    # 3C 345 (comparison with Vilppu reduction and CAFOS observation)

    res = qs_res.filter(instrument=INSTRUMENTS.DIPOL, epoch__night="2023-10-09").get(astrosource__name="1641+399")
    assert res.p == approx(30.1/100, abs=max(2*res.p_err, 0.5/100)) 
    assert res.chi == approx(45.5, abs=max(2*res.chi_err, 5))

@pytest.mark.django_db(transaction=True)
def test_astrometric_calibration(load_test_catalog):

    from iop4lib.db import Epoch, RawFit, ReducedFit, AstroSource
    from iop4lib.enums import IMGTYPES, SRCTYPES, INSTRUMENTS
    from iop4lib.utils.quadmatching import distance

    # get epochs that have DIPOL observations in them
    from iop4lib.iop4 import list_local_epochnames
    epochname_L = list_local_epochnames()
    epoch_L = [Epoch.create(epochname=epochname) for epochname in epochname_L]
    epoch_L = [epoch for epoch in epoch_L if epoch.rawfits.filter(instrument=INSTRUMENTS.DIPOL).exists()]

    for epoch in epoch_L:
        epoch.build_master_biases()

    for epoch in epoch_L:
        epoch.build_master_darks()

    for epoch in epoch_L:
        epoch.build_master_flats()


    # Test 1. Photometry fields

    fileloc = "OSN-T090/2023-11-06/BLLac_IAR-0001R.fit"
    rawfit = RawFit.by_fileloc(fileloc=fileloc)
    redf = ReducedFit.create(rawfit=rawfit)
    redf.build_file()

    fileloc = "OSN-T090/2023-11-13/OJ248_R_full_IAR-0001R.fit"
    rawfit = RawFit.by_fileloc(fileloc=fileloc)
    redf = ReducedFit.create(rawfit=rawfit)
    redf.build_file()

    # Test 2. Polarimetry field using quad matching (I)

    fileloc = "OSN-T090/2023-11-06/BLLAC_R_IAR-0760.fts"
    rawfit = RawFit.by_fileloc(fileloc=fileloc)
    redf = ReducedFit.create(rawfit=rawfit)
    redf.build_file()

    # check source position in the image

    src = AstroSource.objects.get(name="2200+420")

    assert redf.header_hintobject.name == src.name
    assert redf.sources_in_field.filter(name=src.name).exists()
    
    pos_O = src.coord.to_pixel(wcs=redf.wcs1)
    pos_E = src.coord.to_pixel(wcs=redf.wcs2)

    assert (distance(pos_O, [634, 297]) < 25) # O position
    assert (distance(pos_E, [437, 319]) < 50) # E position # might be worse b.c. of companion star

    # Test 3. Polarimetry field using quad matching (II)

    fileloc = "OSN-T090/2023-10-11/OJ248_R_IAR-0111.fts"
    rawfit = RawFit.by_fileloc(fileloc=fileloc)
    redf = ReducedFit.create(rawfit=rawfit)
    redf.build_file()

    # check source position in the image

    src = AstroSource.objects.get(name="0827+243")
    
    assert redf.header_hintobject.name == src.name
    assert redf.sources_in_field.filter(name=src.name).exists()
    
    pos_O = src.coord.to_pixel(wcs=redf.wcs1)
    pos_E = src.coord.to_pixel(wcs=redf.wcs2)

    assert (distance(pos_O, [618, 259]) < 50) # O position
    assert (distance(pos_E, [402, 268]) < 50) # E position

    # Test 4. Polarimetry field using target E, O
    
    fileloc = "OSN-T090/2023-10-25/HD204827_R_IAR-0384.fts"
    rawfit = RawFit.by_fileloc(fileloc=fileloc)
    redf = ReducedFit.create(rawfit=rawfit)
    redf.build_file()

    # check source position in the image

    src = AstroSource.objects.get(name="HD 204827")

    assert redf.header_hintobject.name == src.name
    assert redf.sources_in_field.filter(name=src.name).exists()
    
    pos_O = src.coord.to_pixel(wcs=redf.wcs1)
    pos_E = src.coord.to_pixel(wcs=redf.wcs2)

    assert (distance(pos_O, [684, 397]) < 50) # O position
    assert (distance(pos_E, [475, 411]) < 50) # E position




def test_quad_ish():
    r""" Test that the quad-ish matching works as expected. Procedure:
        1. Create N random 4-point sets
           a) Check if quad coords are invariant under permutations of pts
           b) Check that hash is invariant under permutations of pts
        2. Apply a random translation + rotation + reflection (X and/or Y) to them
           a) Check that the hashes for them are the equal.
        3. Check that the qorder_ish function works, i.e. that the order of the points is the same. We do this by:
           a) randomly ordering transformed points
           b) transforming them back
           c) checking the positions are the same to the original ones, one by one.
    """
    from iop4lib.utils.quadmatching import hash_ish, qorder_ish, quad_coords_ish

    N = 1000 

    # 1.a Create N random 4-point sets

    logger.debug("1. Creating random 4-point sets")

    list_of_points = list()
    for i in range(N):
        list_of_points.append(np.random.rand(4,2))

    # 1.a Check if quad coords are invariant under permutations of pts

    logger.debug("1.a. Checking if quad coords are invariant under permutations of pts")

    for points in list_of_points:
        assert np.allclose(quad_coords_ish(*points)[0], quad_coords_ish(*points[np.random.permutation(4),:])[0])

    # 1.b Check that hash is invariant under permutations of pts

    logger.debug("1.b. Checking that hash is invariant under permutations of pts")

    for points in list_of_points:
        assert hash_ish(points) == approx(hash_ish(points[np.random.permutation(4),:]))

    # 2. Apply a random translation + rotation + reflection to them

    logger.debug("2. Applying random transformations")

    list_of_points_transformed = list()
    M_L = list()
    t_L = list()
    for points in list_of_points:

        # translation
        t = np.random.rand(2)

        # orthogonal transformation (proper or improper rotation)
        from scipy.linalg import qr
        Q, R = qr(np.random.rand(2,2))
        M = Q.dot(Q.T)
        points = M@points.T + t[:,None]

        list_of_points_transformed.append(points.T)
        M_L.append(M)
        t_L.append(t)

    # 2.a Check that the hashes for them are the equal.

    logger.debug("2.a. Checking that the hashes are equal points")

    for p1, p2 in zip(list_of_points, list_of_points_transformed):
        logger.debug(f"p1 = {p1}, p2 = {p2}")
        assert hash_ish(p1) == approx(hash_ish(p2))

    # 3. Check that the qorder_ish function works

    logger.debug("3. Checking that the qorder_ish function works")

    # 3.a) randomly ordering transformed points

    logger.debug("3.a. Randomly ordering transformed points")

    list_of_points_transformed_reordered = list()
    for points in list_of_points_transformed:
        list_of_points_transformed_reordered.append(points[np.random.permutation(4),:])

    # 3.b) transforming them back (substract the translation, apply the inverse of the orthogonal transformation)

    logger.debug("3.b. Transforming them back")

    list_of_points_transformed_reordered_back = list()
    for points, M, t in zip(list_of_points_transformed_reordered, M_L, t_L):
        points = points - np.repeat(t[None,:], 4, axis=0)
        points = np.linalg.inv(M)@points.T
        list_of_points_transformed_reordered_back.append(points.T)

    # 3.c) checking the positions are the same to the original ones, one by one.

    logger.debug("3.c. Checking the positions are the same to the original ones, one by one.")

    for p1, p2 in zip(list_of_points, list_of_points_transformed_reordered_back):

        p1_ordered = np.array(qorder_ish([p for p in p1])).T
        p2_ordered = np.array(qorder_ish([p for p in p2])).T
        
        assert np.all(p1_ordered == approx(p2_ordered))

    

    

    


from django.contrib import admin

from django.utils.html import format_html
from django.urls import path, reverse 
from django.http import HttpResponse, HttpResponseRedirect, HttpResponseNotFound
from django.shortcuts import render
from django.db.models import Q

# other imports
from iop4api.filters import *
from iop4api.models import *
from iop4lib.enums import *
from .fitfile import AdminFitFile

# logging
import logging
logger = logging.getLogger(__name__)

class AdminReducedFit(AdminFitFile):
    model = ReducedFit
    list_display = ["id", 'filename', 'telescope', 'night', 'instrument', 'status', 'imgtype', 'imgsize', 'band', 'obsmode', 'rotangle', 'exptime', 'get_targets_in_field', 'juliandate', 'options', 'modified']
    readonly_fields = [field.name for field in ReducedFit._meta.fields]
    search_fields = ['id', 'filename', 'epoch__telescope', 'epoch__night', 'sources_in_field__name']
    ordering = ['-epoch__night', '-epoch__telescope', '-juliandate']
    list_filter = (
        RawFitIdFilter,
        RawFitTelescopeFilter,
        RawFitNightFilter,
        RawFitFilenameFilter,
        RawFitInstrumentFilter,
        RawFitFlagFilter,
        "obsmode",
        "band",
        "imgsize",
    )
    
    @admin.display(description='OPTIONS')
    def options(self, obj):
        url_rawfit = reverse('iop4admin:%s_%s_changelist' % (RawFit._meta.app_label, RawFit._meta.model_name)) + f"?id={obj.rawfit.id}"
        url_details = reverse('iop4admin:iop4api_reducedfit_details', args=[obj.id])
        url_viewer= reverse('iop4admin:iop4api_reducedfit_viewer', args=[obj.id])
        return format_html(rf'<a href="{url_rawfit}">raw</a> / <a href="{url_details}">details</a> / <a href="{url_viewer}">advanced viewer</a>')
    
    @admin.display(description='IMGSIZE')
    def imgsize(self, obj):
        return obj.imgsize
    
    @admin.display(description='BAND')
    def band(self, obj):
        return obj.band

    @admin.display(description='OBSMODE')
    def obsmode(self, obj):
        return OBSMODES(obj.obsmode).label
    
    @admin.display(description='ROTANGLE')
    def rotangle(self, obj):
        return obj.rotangle
    
    @admin.display(description='EXPTIME')
    def exptime(self, obj):
        return obj.exptime
    
    @admin.display(description='SRCS IN FIELD')
    def get_targets_in_field(self, obj):

        cat_targets = list(obj.sources_in_field.filter(Q(srctype=SRCTYPES.BLAZAR) | Q(srctype=SRCTYPES.STAR)).values_list('name', flat=True))

        if len(cat_targets) > 0:
            return cat_targets
        
        try:
            kw_obj_val = obj.rawfit.header['OBJECT']
        except FileNotFoundError:
            return format_html(f"<i>rawfit not found</i>")
        
        guessed_target = AstroSource.objects.filter(Q(name__icontains=kw_obj_val) | Q(other_names__icontains=kw_obj_val)).values_list('name', flat=True)

        if len(guessed_target) > 0:
            return format_html(f"<i>guess: {guessed_target[0]} ?</i>") 
        
        return format_html(f"<i>kw: '{kw_obj_val}' ?</i>")


# %% [markdown]
# # Database queries

# Note: this notebooks makes reference to other example notebooks, make sure that the following files are available in the same directory as this notebook:

# %%
# !ls *.py

# %% [markdown]
# Configure IOP4 and logging (see examples)
# %%
# %autoawait off
# %matplotlib inline
# %run 01_notebook_configuration.py

# %% [markdown]
# In IOP4, observing sessions, FITS files and observational results are all 
# represented by a Python class, and each instance of them is a row in the 
# corresponding table in the database. IOP4 uses the Django ORM to interact with
# the database (DB) through these models, so the user does not need to worry about
# keeping the results in sync with the database, building SQL queries or 
# updating the DB schema after each change in any of these classes.

# IOP4 models are available under the iop4lib.db submodule. The following models
# are available:
# - Epoch: an observing session, represented by the telescope and date 
# (of the night). It provides methods for downloading and making the initial 
# classification of files from each observing session.
# - RawFit: a raw FITS file
# - MasterBias, MasterDark and MasterDark: calibration frames built from the 
# corresponding raw files for each epoch.
# - ReducedFit: a reduced FITS file, built from the corresponding raw file after 
# applying calibration frames, astronometric calibration, etc.
# - AperPhotResult: result of performing aperture photometry on an image. It is 
# used as an intermediate object to compute photo-polarimetry because of two 
# reasons: to re-use them and to easy debugging of the final results.
# - PhotoPolResult: final result of photo-polarimetry.

# Each of the models is linked through attributes to the rest of models.

# Database queries can be made through the `.objects` attribute of this models, 
# which allows for filtering, ordering, etc of these.

# Let's query the DB for the last observing session:

# %%
from iop4lib.enums import *
from iop4lib.db import *

epoch = Epoch.objects.order_by('-night').first()
epoch

# %% [markdown]
# In an IPython terminal, this will pretty-print a description of `epoch`.
# To query the last of the "science" (`LIGHT`) CAFOS files, we could do

# %%
rf = RawFit.objects.filter(imgtype=IMGTYPES.LIGHT, instrument=INSTRUMENTS.CAFOS).order_by('-juliandate').first()
rf

# %% [markdown]
# If the image was already reduced, you can get the corresponding reduced FITS 
# with

# %%
redf = rf.reduced
redf

# %% [markdown]
# If you wanted to get a preview of the file, you could plot the 2D data which
# is accesible through the `redf.mdata` attribute. Otherwise, you can use one of
# the utility function in `iop4lib.utils.plotting`.

# %%
from iop4lib.utils.plotting import imshow_w_sources
imshow_w_sources(redf.mdata)

# %% [markdown]
# The reduced fit will link back to the raw file through `redf.rawfit`. It also 
# links to the used calibration frames through `redf.masterbias`, 
# `redf.masterdark`, `redf.masterbias` and `redf.masterdark`, and to the night
# through `redf.epoch`. You can check all related fields for a given model in 
# its [reference](/iop4/docs/iop4lib.html) (fields of class ForeignKey, OneToOneField or 
# ManyToManyField).

# More complex queries can be made using `.filter()` `.exclude()`. These will 
# return a reduced Django Queryset. For example, to query (and count) all files 
# whose filename contains `BLLac` (case in-sensitive)

# %%
qs = ReducedFit.objects.filter(filename__icontains="BLLac")
qs.count()

# %% [markdown]
# The result is a Django QuerySet, which can be further filtered, indexed or 
# converted to a list. To exclude all DIPOL observations you would do:
qs = qs.exclude(instrument=INSTRUMENTS.DIPOL)
qs.count()

# If you wanted to access the second of the returned one you could do
qs[2]

# %% [markdown]
# Epochs (observing nights) are uniquely identified by the telescope and date 
# corresponding to the day of the start of the night (in the DB schema, there 
# is an _unique constaint_ on the `telescope` and `night` fields). Raw FITS 
# files are defined uniquely by their epoch and file name. The convenience 
# functions `Epoch.by_epochname` and `RawFit.by_fileloc" can be used to directly
# search the DB with this string

# %%
Epoch.by_epochname("OSN-T090/2023-11-06").pk == Epoch.objects.get(telescope="OSN-T090", night="2023-11-06").pk

# %% [markdown]
# and

# %%
RawFit.by_fileloc("OSN-T090/2023-11-06/BLLAC_R_IAR-0760.fts").pk == RawFit.objects.get(epoch__telescope="OSN-T090", epoch__night="2023-11-06", filename="BLLAC_R_IAR-0760.fts").pk

# %% [markdown]
# In the previous examples, the difference between `.filter` and `.get` is that 
# the former will return a QuerySet (iterable) which might also be empty, the 
# later will return the single instance matching the query, and raise an 
# exception if there is none or multiple.

# In addition, we have inadvertently used the `field__query` syntax to access 
# methods or related fields. At the beginning we used `filename__icontains`
# instead of simply `filename` to search for a case-insensitive string in the 
# field. And now we have used `epoch__telescope` to access the `telescope` field
# of the related Epoch objects, since RawFit contains simply a reference to 
# `Epoch`, instead of duplicated `telescope` and `night` fields.

# %% [markdown]
# For more information about querying the DB, see the [Django ORM documentation](https://docs.djangoproject.com/en/4.2/topics/db/queries/).


from django.db import models

class IMGTYPES(models.TextChoices):
    """
    Enum for the type of a rawfit file.
    """

    NONE = None, "None"
    ERROR = "ERROR", "Error"
    FLAT = 'FLAT', "Flat"
    DARK = 'DARK', "Dark"
    BIAS = 'BIAS', "Bias"
    LIGHT = 'LIGHT', "Light"

class BANDS(models.TextChoices):
    """
    Enum for the band of a rawfit file.
    """

    NONE = None, "None"
    ERROR = "ERROR", "Error"
    I = "I", "I"
    R = 'R', "R"
    V = 'V', "V"
    U = 'U', "U"
    B = 'B', "B"

class OBSMODES(models.TextChoices):
    """
    Enum for possible observation modes (photometry, polarimetry, etc)
    """

    NONE = None, "None"
    ERROR = "ERROR", "Error"
    PHOTOMETRY = 'PHOTOMETRY', "Photometry"
    POLARIMETRY = 'POLARIMETRY', "Polarimetry"

class INSTRUMENTS(models.TextChoices):
    """
    Enum for possible instruments
    """

    NONE = None, "None"
    CAFOS = 'CAFOS2.2', "CAFOS2.2"
    RoperT90 = 'RoperT90', "RoperT90"
    AndorT90 = 'AndorT90', "AndorT90"
    AndorT150 = 'AndorT150', "AndorT150"
    DIPOL = 'DIPOL', "DIPOL"

class TELESCOPES(models.TextChoices):
    """
    Enum for possible telescopes: (name, name)
    """
    CAHA_T220 = 'CAHA-T220', "CAHA-T220"
    OSN_T090 = 'OSN-T090', "OSN-T090"
    OSN_T150 = 'OSN-T150', "OSN-T150"

class SRCTYPES(models.TextChoices):
    """
    Enum for possible types of sources
    """
    BLAZAR = 'blazar', "Blazar"
    STAR = 'star', "Star"
    TDE = 'tde', "TDE"

class REDUCTIONMETHODS(models.TextChoices):
    """
    Enum for possible reduction methods
    """
    RELPHOT = 'relphot', "Relative photometry"
    RELPOL = 'relpol', "Relative photo-polarimetry"
    ABSPHOT = 'absphot', "Absolute photometry"

class PAIRS(models.TextChoices):
    O = "O", "Ordinary"
    E = "E", "Extraordinary"


import pytest
from pathlib import Path

from .conftest import TEST_CONFIG

# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_path=TEST_CONFIG)

# other imports
import os
from pytest import approx

# logging
import logging
logger = logging.getLogger(__name__)

# fixtures
from .fixtures import load_test_catalog


@pytest.mark.django_db(transaction=True)
def test_testdata(testdata):
    """Test that the test data is available"""
    assert (os.path.exists(Path(iop4conf.datadir) / "raw" / "OSN-T090"))


@pytest.mark.django_db(transaction=True)
def test_testconfig_testdb(load_test_catalog):
    """ Check that the DB is clean (it should be the test database), if it is not, all test will fail """
    from iop4lib.db import Epoch, RawFit, ReducedFit, MasterBias, MasterFlat, AperPhotResult, PhotoPolResult, AstroSource
    assert (hasattr(iop4conf, "basedir"))
    assert (hasattr(iop4conf, "datadir"))
    assert (hasattr(iop4conf, "db_path"))
    assert (hasattr(iop4conf, "config_path"))
    assert (Path(iop4conf.datadir).name == "iop4testdata")
    assert (Path(iop4conf.config_path).name == "config.tests.yaml")
    assert ("test" in Path(iop4conf.db_path).name)
    assert (Epoch.objects.count() == 0)
    assert (RawFit.objects.count() == 0)
    assert (ReducedFit.objects.count() == 0)
    assert (MasterBias.objects.count() == 0)
    assert (MasterFlat.objects.count() == 0)
    assert (AperPhotResult.objects.count() == 0)
    assert (PhotoPolResult.objects.count() == 0)

    # there should be some test sources in the DB, and their calibrators
    assert AstroSource.objects.count() > 0
    assert AstroSource.objects.filter(name="2200+420").exists()


@pytest.mark.django_db(transaction=True)
def test_host_correction_data_load(load_test_catalog):
    r""" Check that the host correction data can be correctly loaded """
    from iop4lib.db import AstroSource
    from iop4lib.utils import get_host_correction

    val, err = get_host_correction(AstroSource.objects.get(name="2200+420"), 5)

    assert val is not None
    assert val == approx(1.18)
    assert err == approx(0.06)

@pytest.mark.django_db(transaction=True)
def test_iop4_script():
    """ Test that the iop4 script is available """
    import subprocess
    assert subprocess.run(["iop4", "--help"]).returncode == 0, "iop4 help invokation"
    assert subprocess.run(["iop4", "--check-config"]).returncode == 0, "iop4 configuration check"


# other imports
import os, yaml, logging
from pathlib import Path
from importlib import resources
import matplotlib, matplotlib.pyplot

# Disable matplotlib logging except for warnings and above

matplotlib.pyplot.set_loglevel('warning')

# logging
import logging
logger = logging.getLogger(__name__)

# Set journal_mode=WAL and synchronous=NORMAL for sqlite3 databases on 
# connection, to improve support for concurrent write access to the 
# database during parallel reduction

from django.db.backends.signals import connection_created
from django.dispatch.dispatcher import receiver

@receiver(connection_created)    
def enable_wal_sqlite(sender, connection, **kwargs) -> None:
    if connection.vendor == "sqlite":
        with connection.cursor() as cursor:
            cursor.execute('PRAGMA synchronous=NORMAL;')
            cursor.execute('PRAGMA journal_mode=WAL;')

class Config(dict):
    r""" Configuration class for IOP4.
    
    This class is a singleton, that is, it always returns the same instance of the class.
    """

    # PATH CONFIG

    basedir = Path(__file__).parents[1].absolute()

    # ALL OTHER CONFIG OPTIONS READ FROM SEPARATE CONFIG FILE (see config.example.yaml)

    _instance = None
    _configured = False

    def __new__(cls, *args, **kwargs):
        """ returns always the same instance of the class. """
        if Config._instance is None:
            Config._instance =  super().__new__(cls)
        return Config._instance
        
    def __init__(self, reconfigure=False, **kwargs):
        """
        Reads the specified config file, or the default one, and creates a configuration object.
        
        The configuration is performed only once, the first time it is called, unless you pass reconfigure=True.

        Parameters
        ----------
        config_path: str, default None
            Path to the configuration file. If None, it will try to use, in order: 
            1. The current config file
            2. The file specified by the environment variable IOP4_CONFIG_FILE 
            3. ~/.iop4.config.yaml, if it exists
            4. The example config file in the iop4lib package.
        config_db : bool, default False
            If True, configures django ORM to make the models and database accesible. It 
            should be used only once the top level of your script.
        gonogui : bool, default True
            If True, configures matplotlib to work without a GUI.
        jupytermode : bool, default False
            Code in jupyter notebooks is actually executed in an async context, which django ORM tries to
            protect against. This async context should in principle be disabled running '%autoawait off' in
            jupyter, but this might fail.

            If jupytermode is True, iop4lib will configure django ORM to disable this protection (setting
            the environment variable DJANGO_ALLOW_ASYNC_UNSAFE to true). You should try first to run '%autoawait off'.

            For IPython, you should try also '%autowait off' first too. Moreover, it seems that is not necessary to 
            use jupytermode=True in IPython Shell, even without running '%autowait off'.

        Returns
        -------
        self

        Attributes
        ----------
        Different configuration parameters such as database path, log format, credentials 
        for data download from remote sources, etc.
        """

        super(Config, self).__init__(**kwargs)
        self.__dict__ = self

        if reconfigure or not Config._configured:
            self.configure(**kwargs)

    def copy(self, iop4conf):
        for k, v in iop4conf.items():
            setattr(self, k, v)

    def configure(self, config_path=None, config_db=False, config_logging=True, gonogui=True, jupytermode=False, **kwargs):
    
        Config._configured = True

        # If config_path is None, either use the one already in use or the default one

        if config_path is None:
            if hasattr(self, 'config_path') and self.config_path is not None:
                config_path = self.config_path
            else:

                if os.getenv("IOP4_CONFIG_FILE") is not None:
                    config_path = Path(os.getenv("IOP4_CONFIG_FILE")).expanduser()
                elif Path("~/.iop4.config.yaml").expanduser().exists():
                    config_path = Path("~/.iop4.config.yaml").expanduser()
                else:
                    config_path = resources.files("iop4lib") / "config.example.yaml"

                if not config_path.exists():
                    raise FileNotFoundError(f"Config file {config_path} not found.")
        
        self.config_path = Path(config_path).expanduser()

        # Load config file and set attributes

        with open(config_path, 'r') as f:
            config_dict = yaml.safe_load(f)

        for k, v in config_dict.items():
            setattr(self, k, v)

        # Override with config options in the environment
        for k, v in os.environ.items():
            if k.startswith("IOP4_") and k != "IOP4_CONFIG_FILE":
                setattr(self, k[5:].lower(), v)

        # Override with config options passed as kwargs

        for k, v in kwargs.items():
            setattr(self, k, v)

        # Allow paths relative to home directory

        for k, v in self.items():
            if (k in ['basedir', 'datadir', 'log_file'] or 'path' in k) and v is not None: # config variables should have ~ expanded
                setattr(self, k, str(Path(v).expanduser()))

        # If data dir does not exist, create it
        
        if not os.path.exists(self.datadir):
            os.makedirs(self.datadir)

        # Check if the logs subdirectory exists, if not, create it
        
        if not os.path.exists(os.path.join(self.datadir, "logs")):
            os.makedirs(os.path.join(self.datadir, "logs"))

        # Load OSN names from external file if indicated, load them into patterns like name*.fit, name*.fits, or name*.fts.

        if self.osn_source_list_path is not None and os.path.exists(self.osn_source_list_path):
            with open(self.osn_source_list_path, 'r') as f:
                self.osn_fnames_patterns += [fr"(^{s[:-1]}.*\.fi?ts?$)" for s in f.readlines() if s[0] != '#']

        if config_logging:
            self.configure_logging()
            
        if gonogui:
            matplotlib.use("Agg")

        if jupytermode:
            os.environ["DJANGO_ALLOW_ASYNC_UNSAFE"] = "true"

        if config_db:
            self.configure_db()

    def configure_db(self):

        import django
        from django.conf import settings
        from django.apps import apps
        
        settings.configure(
            INSTALLED_APPS=[
                'iop4api',
            ],
            DATABASES = {
                "default": {
                    "ENGINE": "django.db.backends.sqlite3",
                    "NAME": self.db_path,
                }
            },
            DEBUG = False,
            LOGGING_CONFIG = None, # otherwise breaks logging colors inside docs notebooks
            TEMPLATES = [ # template engine needed for the automatic summary emails
                {
                    'BACKEND': 'django.template.backends.django.DjangoTemplates',
                    'DIRS': [],
                },
            ],
        )
        
        apps.populate(settings.INSTALLED_APPS)
    
        django.setup()


    def configure_logging(self):
        r""" Adds the '%(proc_memory)' field in the log record."""
        old_factory = logging.getLogRecordFactory()

        def record_factory_w_proc_memory(*args, **kwargs):
            from iop4lib.utils import get_mem_current
            record = old_factory(*args, **kwargs)
            record.proc_memory = f"{get_mem_current()/1024**2:.0f} MB"
            return record

        if old_factory.__name__ == record_factory_w_proc_memory.__name__:
            return
        
        logging.setLogRecordFactory(record_factory_w_proc_memory)


    def is_valid(self):
        r""" Checks that the configuration file is correct by comparing it with the default one.
        
        Returns
        -------
        bool
            True if the configuration file is correct, False otherwise.

        """

        with open(self.config_path, 'r') as f:
            config_dict = yaml.safe_load(f)

        with open(resources.files("iop4lib") / "config.example.yaml", 'r') as f:
            config_dict_example = yaml.safe_load(f)

        wrong = False

        for k, v in config_dict_example.items():
            if k not in config_dict:
                logger.error(f"ERROR: {k} missing in config file.")
                wrong = True

        return not wrong


# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

# django imports
from django.contrib import admin, messages
from django.utils.html import format_html
from django.urls import path, reverse 
from django.db.models import Avg

# other imports
from iop4api.models import AstroSource, ReducedFit
from iop4admin import views

import numpy as np
from astroquery.mast import Catalogs
from astropy.coordinates import SkyCoord

# iop4lib imports
from iop4lib.enums import BANDS

# logging
import logging
logger = logging.getLogger(__name__)

class AdminAstroSource(admin.ModelAdmin):
    model = AstroSource
    list_display = ['name', 'other_names', 'ra_hms', 'dec_dms', 'srctype', 'get_last_reducedfit', 
                    'get_last_mag_R', 'get_calibrates',
                    'get_texp_andor90', 'get_texp_andor150', 'get_texp_dipol', 
                    'get_comment_firstline', 'get_details']
    search_fields = ['name', 'other_names', 'ra_hms', 'dec_dms', 'srctype', 'comment']
    list_filter = ('srctype',)
    actions = ['add_field_stars_from_panstarrs', 'remove_field_stars_from_panstarrs']
    
    @admin.display(description='CALIBRATES')
    def get_calibrates(self, obj):
        stars_str_L = obj.calibrates.all().values_list('name', flat=True)
        return "\n".join(stars_str_L)
    
    @admin.display(description='COMMENTS')
    def get_comment_firstline(self, obj):
        lines = obj.comment.split("\n")
        txt = lines[0]
        if len(lines) > 0 or len(lines[0]) > 30:
            txt = txt[:30] + "..."
        return txt
    
    @admin.display(description='LAST FILE')
    def get_last_reducedfit(self, obj):
        redf = obj.last_reducedfit
        if redf is not None:
            url = reverse('iop4admin:%s_%s_changelist' % (ReducedFit._meta.app_label, ReducedFit._meta.model_name)) + "?id=%s" % redf.pk
            return format_html(rf'<a href="{url}">{redf.epoch.night}</a>')
        else:
            return None
    
    @admin.display(description="LAST MAG")
    def get_last_mag_R(self, obj):
        mag_r_avg, mag_r_err_avg = obj.last_night_mag_R
        if mag_r_avg is not None:
            return f"{mag_r_avg:.2f}"
        else:
            return None
    
    @admin.display(description='DETAILS')
    def get_details(self, obj):
        url = reverse('iop4admin:iop4api_astrosource_details', args=[obj.id])
        return format_html(rf'<a href="{url}">Details</a>')
    
    def get_urls(self):
        urls = super().get_urls()
        my_urls = [
            path('details/<int:pk>', self.admin_site.admin_view(views.AstroSourceDetailsView.as_view()), name='iop4api_astrosource_details'),
        ]
        return my_urls + urls
    
    @admin.display(description='T. Andor90')
    def get_texp_andor90(self, obj):
        last_night_mag_R, _ = obj.last_night_mag_R
        texp = obj.texp_andor90

        if last_night_mag_R is None:
            return None
        
        if texp is None:
            return "X"

        return texp
    
    @admin.display(description='T. Andor150')
    def get_texp_andor150(self, obj):
        last_night_mag_R, _ = obj.last_night_mag_R
        texp = obj.texp_andor150

        if last_night_mag_R is None:
            return None
        
        if texp is None:
            return "X"

        return texp
    
    @admin.display(description='T. x N DIPOL')
    def get_texp_dipol(self, obj):
        last_night_mag_R, _ = obj.last_night_mag_R
        texp = obj.texp_dipol
        nreps = obj.nreps_dipol

        if last_night_mag_R is None:
            return None
        
        if texp is None:
            return "X"
        
        return f"{texp} x {nreps}"

    @admin.action(description='Automatically add field stars from PanSTARRS')
    def add_field_stars_from_panstarrs(self, request, queryset):

        if queryset.count() > 1:
            messages.error(request, "Only one source at a time, please.")
            return
        
        main_src = queryset.first()

        # quality constraints
        MIN_R_MAG = 11 # minimum R magnitude
        MAX_R_MAG = 16  # maximum R magnitude
        MAX_MAG_STD = 0.02 # maximum standard deviation in the required bands (we want our calibrators to be stable)
        MIN_N_OBS = 5 # minimum number of observations in the required bands

        logger.info(f"Querying PanSTARRS around {main_src.name} ({main_src.coord.ra.deg} {main_src.coord.dec.deg})")

        try:
            catalog_data = Catalogs.query_criteria(coordinates=f"{main_src.coord.ra.deg} {main_src.coord.dec.deg}", catalog="PANSTARRS", version=1, radius="0.12 deg")
        except Exception as e:
            logger.error(f"Error querying PanSTARRS around {main_src.name} ({main_src.coord.ra.deg} {main_src.coord.dec.deg}): {e}")
            messages.error(request, f"Error querying PanSTARRS around {main_src.name}: {e}")
            return

        logger.info(f"Got {len(catalog_data)} PanSTARRS field stars around {main_src.name}")

        column_names = ['objName', 'raMean', 'decMean',
                        'rMeanApMag', 'rMeanApMagErr', 'rMeanApMagStd', 'rMeanApMagNpt',
                        'gMeanApMag', 'gMeanApMagErr', 'gMeanApMagStd', 'gMeanApMagNpt',
                        'iMeanApMag', 'iMeanApMagErr', 'iMeanApMagStd', 'iMeanApMagNpt',
                        'yMeanApMag', 'yMeanApMagErr', 'yMeanApMagStd', 'yMeanApMagNpt']
        
        for col in column_names:

            if np.issubdtype(catalog_data.dtype[col], np.number):
                idx = np.isfinite(catalog_data[col])
                idx = idx.data & idx.mask
                catalog_data = catalog_data[~idx]

        idx = (MIN_R_MAG <= catalog_data["rMeanApMag"]) & (catalog_data["rMeanApMag"] <= MAX_R_MAG)
        idx = idx & (catalog_data["rMeanApMagStd"] < MAX_MAG_STD)
        idx = idx & (catalog_data["rMeanApMagNpt"] > MIN_N_OBS)
        idx = idx & (catalog_data["gMeanApMagStd"] < MAX_MAG_STD)
        idx = idx & (catalog_data["gMeanApMagNpt"] > MIN_N_OBS)
        idx = idx & (catalog_data["iMeanApMagStd"] < MAX_MAG_STD)
        idx = idx & (catalog_data["iMeanApMagNpt"] > MIN_N_OBS)
        idx = idx & (catalog_data["yMeanApMagStd"] < MAX_MAG_STD)
        idx = idx & (catalog_data["yMeanApMagNpt"] > MIN_N_OBS)
        catalog_data = catalog_data[idx]

        catalog_data.sort(['rMeanApMagStd', 'rMeanApMagErr'])

        catalog_data.remove_columns([col for col in catalog_data.columns if col not in column_names])

        logger.info(f"Filtered down to {len(catalog_data)} PanSTARRS field stars with {MIN_R_MAG} <= R <= {MAX_R_MAG}, std < {MAX_MAG_STD} and n_obs > {MIN_N_OBS}")

        if len(catalog_data) == 0:
            logger.error(f"No PanSTARRS field stars found for {main_src.name}, skipping")

        # sort by number of observations in R, take top 10 only
        if len(catalog_data) > 10:
            logger.info("Keeping only top 10 field stars by number of R observations")
            catalog_data.sort('rMeanApMagNpt')
            catalog_data = catalog_data[-10:]

        field_stars = list()

        for row in catalog_data:
            try:
                cat_ra, cat_dec = row["raMean"], row["decMean"]
                ra_hms, dec_dms = SkyCoord(cat_ra, cat_dec, unit="deg").to_string("hmsdms").split()

                # Transform SDSS to Johnson-Cousins (Lupton 2005)

                B = row["gMeanApMag"] + 0.3130*(row["gMeanApMag"] - row["rMeanApMag"]) + 0.2271 
                err_B = np.sqrt((1.313*row["gMeanApMagErr"])**2+(0.313*row["rMeanApMagErr"])**2+0.0107**2)

                V = row["gMeanApMag"] - 0.5784*(row["gMeanApMag"] - row["rMeanApMag"]) - 0.0038  
                err_V = np.sqrt((0.4216*row["gMeanApMagErr"])**2+(0.5784*row["rMeanApMagErr"])**2+0.0054**2)

                R = row["rMeanApMag"] - 0.2936*(row["rMeanApMag"] - row["iMeanApMag"]) - 0.1439  
                err_R = np.sqrt((0.7064*row["rMeanApMagErr"])**2+(0.2936*row["iMeanApMagErr"])**2+0.0072**2)

                I = row["rMeanApMag"] - 1.2444*(row["rMeanApMag"] - row["iMeanApMag"]) - 0.3820;  
                err_I = np.sqrt((0.2444*row["rMeanApMagErr"])**2+(1.2444*row["iMeanApMagErr"])**2+0.0078**2)

                field_star = AstroSource(name=f"PanSTARRS {cat_ra:.5f} {cat_dec:.5f}", 
                                        other_names=row["objName"],
                                        ra_hms=ra_hms, dec_dms=dec_dms,
                                        comment=(f"PanSTARRS field star for {main_src.name}.\n"
                                                f"Object name: `{row['objName']}`\n"
                                                f"Autogenerated from PanSTARRS catalog query.\n"
                                                f"SDSS to Johnson-Cousins transformation by Lupton (2005).\n"
                                                f"Field `other_names` corresponds to PanSTARRS `objName`.\n"),
                                        srctype="star",
                                        mag_B=B, mag_B_err=err_B,
                                        mag_V=V, mag_V_err=err_V,
                                        mag_R=R, mag_R_err=err_R,
                                        mag_I=I, mag_I_err=err_I)
                                                    
                field_star.save()

                field_stars.append(field_star)
            except Exception as e:
                logger.error(f"Error with {row['objName']}: {e}")
                continue

        logger.info(f"Added {len(field_stars)} PanSTARRS field stars for {main_src.name}")

        main_src.calibrators.add(*field_stars)

        messages.success(request, f"Added {len(field_stars)} PanSTARRS field stars for {main_src.name}")


    @admin.action(description='Remove all field stars from PanSTARRS')
    def remove_field_stars_from_panstarrs(self, request, queryset):

        if queryset.count() > 1:
            messages.error(request, "Only one source at a time, please.")
            return
        
        main_src = queryset.first()

        field_stars = main_src.calibrators.filter(name__startswith="PanSTARRS")
        field_stars.delete()
        logger.info(f"Removed {len(field_stars)} PanSTARRS field stars for {main_src.name}")

        messages.success(request, f"Removed {len(field_stars)} PanSTARRS field stars for {main_src.name}")


# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

# django imports
from abc import ABCMeta

# other imports
import astropy.units as u
from astropy.coordinates import Angle, SkyCoord
import astrometry
import numpy as np
import math

# iop4lib imports
from iop4lib.enums import *
from .instrument import Instrument
from iop4lib.telescopes import OSNT090, OSNT150

# logging
import logging
logger = logging.getLogger(__name__)

import typing
if typing.TYPE_CHECKING:
    from iop4lib.db.reducedfit import ReducedFit, RawFit

class OSNCCDCamera(Instrument, metaclass=ABCMeta):
    r""" Abstract class for OSN CCD cameras. """

    required_masters = ['masterbias', 'masterflat']

    @classmethod
    def classify_juliandate_rawfit(cls, rawfit):
        """
        Andor cameras fits has JD keyword
        """
        import astropy.io.fits as fits
        jd = fits.getheader(rawfit.filepath, ext=0)["JD"]
        rawfit.juliandate = jd

    @classmethod
    def classify_imgtype_rawfit(cls, rawfit):
        """
        Andor cameras fits has IMAGETYP keyword: FLAT, BIAS, LIGHT

        .. note::
        **Sometimes, the IMAGETYP keyword is wrong**, it has LIGHT on it but the filename and the OBJECT keyword contain the word "Flat". In those ocassions, it should be classified as
        a FLAT.
        """
        from iop4lib.db.rawfit import RawFit
        import astropy.io.fits as fits

        with fits.open(rawfit.filepath) as hdul:
            if hdul[0].header['IMAGETYP'] == 'FLAT':
                rawfit.imgtype = IMGTYPES.FLAT
            elif hdul[0].header['IMAGETYP'] == 'BIAS':
                rawfit.imgtype = IMGTYPES.BIAS
            elif hdul[0].header['IMAGETYP'] == 'LIGHT':
                rawfit.imgtype = IMGTYPES.LIGHT
                # workarounds for wrong keyword in OSN (see note in docstring)
                if "FLAT" in hdul[0].header["OBJECT"].upper() and "FLAT" in rawfit.filename.upper():
                    rawfit.imgtype = IMGTYPES.FLAT
                elif "BIAS" in hdul[0].header["OBJECT"].upper() and "BIAS" in rawfit.filename.upper():
                    rawfit.imgtype = IMGTYPES.BIAS
            else:
                logger.error(f"Unknown image type for {rawfit.fileloc}.")
                rawfit.imgtype = IMGTYPES.ERROR
                raise ValueError

    @classmethod
    def classify_band_rawfit(cls, rawfit):
        """
            OSN Files have no FILTER keyword if they are BIAS, FILTER=Clear if they are FLAT, and FILTER=FilterName if they are LIGHT.
            For our DB, we have R, U, ..., None, ERROR.

            For polarimetry, which is done by taking four images with the R filter at different angles, we have R_45, R0, R45, R90.
        """

        from iop4lib.db.rawfit import RawFit

        if 'FILTER' not in rawfit.header:
            if rawfit.imgtype == IMGTYPES.BIAS:
                rawfit.band = BANDS.NONE
            else:
                rawfit.band = BANDS.ERROR
                raise ValueError(f"Missing FILTER keyword for {rawfit.fileloc} which is not a bias (it is a {rawfit.imgtype}).")
        elif rawfit.header['FILTER'] in {"Clear", ""}:  
            if rawfit.imgtype == IMGTYPES.FLAT:
                rawfit.band = BANDS.NONE
            else:
                rawfit.band = BANDS.ERROR
                raise ValueError(f"FILTER keyword is 'Clear' for {rawfit.fileloc} which is not a flat (it is a {rawfit.imgtype}).")
        else:
            rawfit.band = rawfit.header['FILTER'][0] # First letter of the filter name (R, U, ...) includes cases as R45, R_45, etc   

    @classmethod
    def classify_obsmode_rawfit(cls, rawfit):
        """
        In OSN Andor Polarimetry, we only have polarimetry for filter R, and it is indicated as R_45, R0, R45, R90 (-45, 0, 45 and 90 degrees). They correspond
        to the different angles of the polarimeter.

        For photometry, the filter keyword willl be simply the letter R, U, etc.

        The values for angles are -45, 0, 45 and 90.

        Lately we have seen "R-45" instead of "R_45", so we have to take care of that too.
        """

        from iop4lib.db.rawfit import RawFit
        import re

        if rawfit.band == BANDS.ERROR:
            raise ValueError("Cannot classify obsmode if band is ERROR.")
        
        if rawfit.band == BANDS.R:
            if rawfit.header['FILTER'] == "R":
                rawfit.obsmode = OBSMODES.PHOTOMETRY
            else:
                logger.debug("Band is R, but FILTER is not exactly R, for OSN this must mean it is polarimetry. Trying to extract angle from FILTER keyword.")

                rawfit.obsmode = OBSMODES.POLARIMETRY

                if rawfit.header['FILTER'] == "R_45" or rawfit.header['FILTER'] == "R-45":
                    rawfit.rotangle = -45
                elif rawfit.header['FILTER'] == "R0":
                    rawfit.rotangle = 0
                elif rawfit.header['FILTER'] == "R45" or rawfit.header['FILTER'] == "R+45":
                    rawfit.rotangle = 45
                elif rawfit.header['FILTER'] == "R90":
                    rawfit.rotangle = 90
                else:
                    raise ValueError(f"Cannot extract angle from FILTER keyword '{rawfit.header['FILTER']}'.")
        else:
            logger.debug("Band is not R, assuming it is photometry.")
            rawfit.obsmode = OBSMODES.PHOTOMETRY

    @classmethod
    def get_header_hintcoord(cls, rawfit):
        """ Get the position hint from the fits header as a SkyCoord.

        OSN T090 / AndorT090 have keywords OBJECT, OBJECTRA, OBJECTDEC in the header; e.g: 
        OBJECT 	TXS0506
        OBJCTRA 	05 09 20 ---> this can be input with unit u.hourangle
        OBJCTDEC 	+05 41 16 ---> this can be input with unit u.deg
        """     
        
        hint_coord = SkyCoord(Angle(rawfit.header['OBJCTRA'], unit=u.hourangle), Angle(rawfit.header['OBJCTDEC'], unit=u.degree), frame='icrs')
        return hint_coord
    
    @classmethod
    def get_astrometry_position_hint(cls, rawfit, allsky=False, n_field_width=1.5, hintsep=None):
        """ Get the position hint from the FITS header as an astrometry.PositionHint.

        It will try to use the coordinate/pointing information from the header
        (RA and DEC keywords, or similar), given by `get_header_hintcoord`. If 
        that fails, it will try to use the object information from the header 
        (OBJECT keyword), given by `get_header_hintobject`. If both fail,
        it will raise an exception.

        Parameters
        ----------
            allsky: bool, optional
                If True, the hint will cover the whole sky, and n_field_width and hintsep will be ignored.
            n_field_width: float, optional
                The search radius in units of field width. Default is 1.5.
            hintsep: Quantity, optional
                The search radius in units of degrees.
        """        

        hintcoord = None

        try:
            hintcoord = cls.get_header_hintcoord(rawfit)
        except KeyError:
            logger.error(f"No hint coordinates found in header, trying hint object.")

            try:
                hintobj = cls.get_header_hintobject(rawfit)

                if hintobj is None:
                    logger.exception(f"No hint object found in header (but kw seems to exist).")
                    raise
                
                hintcoord = hintobj.coord
                
            except KeyError:
                logger.exception(f"No hint object found in header for {rawfit.fileloc}.")


        if hintcoord is None:
            raise ValueError(f"Could not build astrometry hint for {rawfit.fileloc}")
        
        if allsky:
            hintsep = 180.0 * u.deg
        else:
            if hintsep is None:
                hintsep = (n_field_width * cls.field_width_arcmin*u.Unit("arcmin"))

        return astrometry.PositionHint(ra_deg=hintcoord.ra.deg, dec_deg=hintcoord.dec.deg, radius_deg=hintsep.to_value("deg"))
    
    @classmethod
    def get_astrometry_size_hint(cls, rawfit):
        """ Get the size hint for this telescope / rawfit.

            According to OSN T090 camera information (https://www.osn.iaa.csic.es/page/camaras-ccdt150-y-ccdt90) 
            the camera pixels are 0.387as/px and it has a field of view of 13,20' x 13,20'. So we provide close values 
            for the hint. If the files are 1x1 it will be 0.387as/px, if 2x2 it will be twice.

            According to OSN T0150 camera information (https://www.osn.iaa.csic.es/page/camaras-ccdt150-y-ccdt90) 
            camera pixels are 0.232as/px and it has a field of view of 7.92' x 7.92'.
            If the files are 1x1 it will be that, if they are 2x2 it will be twice.
        """

        if rawfit.header['NAXIS1'] == 2048:
            return astrometry.SizeHint(lower_arcsec_per_pixel=0.95*cls.arcsec_per_pix, upper_arcsec_per_pixel=1.05*cls.arcsec_per_pix)
        elif rawfit.header['NAXIS1'] == 1024:
            return astrometry.SizeHint(lower_arcsec_per_pixel=2*0.95*cls.arcsec_per_pix, upper_arcsec_per_pixel=2*1.05*cls.arcsec_per_pix)
            
    @classmethod
    def has_pairs(cls, fit_instance):
        return False

    @classmethod
    def compute_relative_polarimetry(cls, polarimetry_group):
        """ Computes the relative polarimetry for a polarimetry group for OSNT090 observations.

        .. note:: 
            The rotation angle in OSNT090 refers to the angle between the polarized filter and some reference direction. This is different
            to the rotation angle for CAHA-T220 which is the angle between the half-wave plate (HW) and its fast (extraordinary) axis. See the docs
            for ``CAHAT220.compute_relative_polarimetry`` for more information.
        
        Instrumental polarization is corrected. Currently values are hardcoded in qoff, uoff, dqoff, duoff, Phi, dPhi (see code),
        but the values without any correction are stored in the DB so the correction can be automatically obtained in the future.
        """
                
        from iop4lib.db.aperphotresult import AperPhotResult
        from iop4lib.db.photopolresult import PhotoPolResult

        logger.debug("Computing %s relative polarimetry for group: %s", cls.name, "".join(map(str,polarimetry_group)))

        # Instrumental polarization

        ## values for T090 TODO: update them  manually or (preferibly) dinamically (TODO)
        ## to compute the instrumental polarization we need to get the mean of the Q and U images, use zero
        ## (done in the _X_nocorr variables)

        qoff = 0.0579
        uoff = 0.0583
        dqoff = 0.003
        duoff = 0.0023
        Phi = math.radians(-18)
        dPhi = math.radians(0.001)

        # Perform some checks on the group

        ## get the band of the group

        bands = [reducedfit.band for reducedfit in polarimetry_group]

        if len(set(bands)) == 1:
            band = bands[0]
        else: # should not happen
            raise Exception(f"Can not compute relative polarimetry for a group with different bands: {bands}")

        ## check obsmodes

        if not all([reducedfit.obsmode == OBSMODES.POLARIMETRY for reducedfit in polarimetry_group]):
            raise Exception(f"This method is only for polarimetry images.")
        
        ## check sources in the fields

        sources_in_field_qs_list = [reducedfit.sources_in_field.all() for reducedfit in polarimetry_group]
        group_sources = set.intersection(*map(set, sources_in_field_qs_list))

        if len(group_sources) == 0:
            logger.error("No common sources in field for all polarimetry groups.")
            return
        
        if group_sources != set.union(*map(set, sources_in_field_qs_list)):
            logger.warning(f"Sources in field do not match for all polarimetry groups: {set.difference(*map(set, sources_in_field_qs_list))}")

        ## check rotation angles

        rot_angles_available = set([redf.rotangle for redf in polarimetry_group])
        rot_angles_required = {0.0, 45.0, 90.0, -45.0}

        if not rot_angles_required.issubset(rot_angles_available):
            logger.error(f"Rotation angles missing: {rot_angles_required - rot_angles_available}; returning early.")
            return

        # 1. Compute all aperture photometries

        aperpix, r_in, r_out, fit_res_dict = cls.estimate_common_apertures(polarimetry_group, reductionmethod=REDUCTIONMETHODS.RELPHOT)
        target_fwhm = fit_res_dict['mean_fwhm']
        
        logger.debug(f"Computing aperture photometries for the {len(polarimetry_group)} reducedfits in the group with target {aperpix:.1f}.")

        for reducedfit in polarimetry_group:
            cls.compute_aperture_photometry(reducedfit, aperpix, r_in, r_out)

        # 2. Compute relative polarimetry for each source (uses the computed aperture photometries)

        logger.debug("Computing relative polarimetry.")

        photopolresult_L = list()

        for astrosource in group_sources:
            
            flux_0 = AperPhotResult.objects.get(reducedfit__in=polarimetry_group, astrosource=astrosource, aperpix=aperpix, pairs="O", reducedfit__rotangle=0.0).flux_counts
            flux_0_err = AperPhotResult.objects.get(reducedfit__in=polarimetry_group, astrosource=astrosource, pairs="O", aperpix=aperpix, reducedfit__rotangle=0.0).flux_counts_err

            flux_45 = AperPhotResult.objects.get(reducedfit__in=polarimetry_group, astrosource=astrosource, pairs="O", aperpix=aperpix, reducedfit__rotangle=45.0).flux_counts
            flux_45_err = AperPhotResult.objects.get(reducedfit__in=polarimetry_group, astrosource=astrosource, pairs="O", aperpix=aperpix, reducedfit__rotangle=45.0).flux_counts_err

            flux_90 = AperPhotResult.objects.get(reducedfit__in=polarimetry_group, astrosource=astrosource, pairs="O", aperpix=aperpix, reducedfit__rotangle=90.0).flux_counts
            flux_90_err = AperPhotResult.objects.get(reducedfit__in=polarimetry_group, astrosource=astrosource, pairs="O", aperpix=aperpix, reducedfit__rotangle=90.0).flux_counts_err

            flux_n45 = AperPhotResult.objects.get(reducedfit__in=polarimetry_group, astrosource=astrosource, pairs="O", aperpix=aperpix, reducedfit__rotangle=-45.0).flux_counts
            flux_n45_err = AperPhotResult.objects.get(reducedfit__in=polarimetry_group, astrosource=astrosource, pairs="O", aperpix=aperpix, reducedfit__rotangle=-45.0).flux_counts_err

            # from IOP3 polarimetry_osn() :

            fluxes = np.array([flux_0, flux_45, flux_90, flux_n45])
            flux_mean = fluxes.mean()
            flux_err = fluxes.std() / math.sqrt(len(fluxes))
            
            qraw = (flux_0 - flux_90) / (flux_0 + flux_90)
            uraw = (flux_45 - flux_n45) / (flux_45 + flux_n45)
            
            #Applying error propagation...
            
            dqraw = qraw * math.sqrt(((flux_0_err**2+flux_90_err**2)/(flux_0+flux_90)**2)+(((flux_0_err**2+flux_90_err**2))/(flux_0-flux_90)**2))
            duraw = uraw * math.sqrt(((flux_45_err**2+flux_n45_err**2)/(flux_45+flux_n45)**2)+(((flux_45_err**2+flux_n45_err**2))/(flux_45-flux_n45)**2))

            qc = qraw - qoff
            uc = uraw - uoff

            dqc = math.sqrt(dqraw**2 + dqoff**2) 
            duc = math.sqrt(duraw**2 + duoff**2)

            q = qc*math.cos(2*Phi) - uc*math.sin(2*Phi)
            u = qc*math.sin(2*Phi) + uc*math.cos(2*Phi)
 
            dqa = qc*math.cos(2*Phi) * math.sqrt((dqc/qc)**2+((2*dPhi*math.sin(2*Phi))/(math.cos(2*Phi)))**2) 
            dqb = uc*math.sin(2*Phi) * math.sqrt((duc/uc)**2+((2*dPhi*math.cos(2*Phi))/(math.sin(2*Phi)))**2)
            dua = qc*math.sin(2*Phi) * math.sqrt((dqc/qc)**2+((2*dPhi*math.cos(2*Phi))/(math.sin(2*Phi)))**2) 
            dub = uc*math.cos(2*Phi) * math.sqrt((duc/uc)**2+((2*dPhi*math.sin(2*Phi))/(math.cos(2*Phi)))**2)
            
            dq = np.sqrt(dqa**2+dqb**2)
            du = np.sqrt(dua**2+dub**2)
            
            P = math.sqrt(q**2 + u**2)
            dP = P * (1/(q**2+u**2)) * math.sqrt((q*dq)**2+(u*du)**2)

            Theta_0 = 0
            Theta = (1/2) * math.degrees(math.atan2(u,q) + Theta_0)
            dTheta = (0.5 * 180.0 / math.pi) * dP/P

            # compute also non-corrected values for computation of instrumental polarization

            _Phi_nocorr = 0 # no rotation correction?
            _qc_nocorr = qraw # no offset correction
            _uc_nocorr = uraw # no offset correction
            _q_nocorr = _qc_nocorr*math.cos(2*_Phi_nocorr) - _uc_nocorr*math.sin(2*_Phi_nocorr)
            _u_nocorr = _qc_nocorr*math.sin(2*_Phi_nocorr) + _uc_nocorr*math.cos(2*_Phi_nocorr) 
            _p_nocorr = math.sqrt(_q_nocorr**2 + _u_nocorr**2)
            _Theta_0_nocorr = 0
            _Theta_nocorr = (1/2) * math.degrees(math.atan2(_u_nocorr,_q_nocorr) + _Theta_0_nocorr)
            _x_px, _y_px = astrosource.coord.to_pixel(polarimetry_group[0].wcs)
            
            # compute instrumental magnitude (same as for CAHA)

            if flux_mean <= 0.0:
                logger.warning(f"{polarimetry_group=}: negative flux mean encountered while relative polarimetry for {astrosource=} ??!! It will be nan, but maybe we should look into this...")

            mag_inst = -2.5 * np.log10(flux_mean) # slower than math.log10 but returns nan when flux < 0 instead of throwing error (see https://github.com/juanep97/iop4/issues/24)
            mag_inst_err = math.fabs(2.5 / math.log(10) * flux_err / flux_mean)

            # if the source is a calibrator, compute also the zero point

            if astrosource.is_calibrator:
                mag_known = getattr(astrosource, f"mag_{band}")
                mag_known_err = getattr(astrosource, f"mag_{band}_err", None) or 0.0

                if mag_known is None:
                    logger.warning(f"Calibrator {astrosource} has no magnitude for band {band}.")
                    mag_zp = np.nan
                    mag_zp_err = np.nan
                else:
                    mag_zp = mag_known - mag_inst
                    mag_zp_err = math.sqrt(mag_known_err ** 2 + mag_inst_err ** 2)
            else:
                mag_zp = None
                mag_zp_err = None

            # save the results
                    
            result = PhotoPolResult.create(reducedfits=polarimetry_group, 
                                           astrosource=astrosource, 
                                           reduction=REDUCTIONMETHODS.RELPOL, 
                                           mag_inst=mag_inst, mag_inst_err=mag_inst_err, mag_zp=mag_zp, mag_zp_err=mag_zp_err,
                                           flux_counts=flux_mean, p=P, p_err=dP, chi=Theta, chi_err=dTheta,
                                           _x_px=_x_px, _y_px=_y_px, _q_nocorr=_q_nocorr, _u_nocorr=_u_nocorr, _p_nocorr=_p_nocorr, _chi_nocorr=_Theta_nocorr,
                                           aperpix=aperpix)
            
            photopolresult_L.append(result)


        # 3. Get average zero point from zp of all calibrators in the group

        calib_mag_zp_array = np.array([result.mag_zp or np.nan for result in photopolresult_L if result.astrosource.is_calibrator]) # else it fills with None also and the dtype becomes object
        calib_mag_zp_array = calib_mag_zp_array[~np.isnan(calib_mag_zp_array)]

        calib_mag_zp_array_err = np.array([result.mag_zp_err or np.nan for result in photopolresult_L if result.astrosource.is_calibrator])
        calib_mag_zp_array_err = calib_mag_zp_array_err[~np.isnan(calib_mag_zp_array_err)]

        if len(calib_mag_zp_array) == 0:
            logger.error(f"Can not compute magnitude during relative photo-polarimetry without any calibrators for this reduced fit.")

        zp_avg = np.nanmean(calib_mag_zp_array)
        zp_std = np.nanstd(calib_mag_zp_array)

        zp_err = np.sqrt(np.nansum(calib_mag_zp_array_err ** 2)) / len(calib_mag_zp_array_err)
        zp_err = math.sqrt(zp_err ** 2 + zp_std ** 2)

        # 4. Compute the calibrated magnitudes for non-calibrators in the group using the averaged zero point

        for result in photopolresult_L:

            if result.astrosource.is_calibrator:
                continue

            result.mag_zp = zp_avg
            result.mag_zp_err = zp_err
        
            result.mag = result.mag_inst + zp_avg
            result.mag_err = math.sqrt(result.mag_inst_err ** 2 + zp_err ** 2)

            result.save()

        # 5. Save results
        for result in photopolresult_L:
            result.save()


class AndorT90(OSNCCDCamera):
    
    name = "AndorT90"
    telescope = OSNT090.name

    instrument_kw_L = ["AndorT90", "RoperT90"] # RoperT90 only in some old images before the PC was updated after camera replacement

    field_width_arcmin = 13.2 
    field_height_arcmin = 13.2
    arcsec_per_pix = 0.387
    gain_e_adu = 4.5

    @classmethod
    def build_wcs(self, reducedfit: 'ReducedFit', *args, **kwargs):
        r""" Overriden for OSN-T090
         
        Overriden to account for HIP2838 U band images with exptime < 2.5, which are known to fail.
        """
        from iop4lib.utils.astrometry import BuildWCSResult

        # OSN-T090 images of HIP2838 in band U with < 2.5s
        # they can not get automatically calibrated because there are almost no sources visible, just return error so we dont loose time trying parameters.

        if reducedfit.band == "U" and "HIP2838" in reducedfit.filename and reducedfit.exptime < 2.5:
            logger.error("Skipping WCS build for HIP2838 U band image with exptime < 2.5 , as it is known to fail, and we will only lose time. Manual calibration is needed for this image. See build_wcs_for_HIP2838_U_band_images.ipynb for more info.")
            return BuildWCSResult(success=False)
        
        return super().build_wcs(reducedfit, *args, **kwargs)

class AndorT150(OSNCCDCamera):
        
    name = "AndorT150"
    telescope = OSNT150.name

    instrument_kw_L = ["Andor", "AndorT150"]

    arcsec_per_pix = 0.232
    gain_e_adu = 4.5
    field_width_arcmin = 7.92
    field_height_arcmin = 7.92


class RoperT90(OSNCCDCamera):
    name = "RoperT90"
    telescope = OSNT090.name

    instrument_kw_L = ["RoperT90"]

    arcsec_per_pix = 0.387
    gain_e_adu = 7.14
    field_width_arcmin = 13.2
    field_height_arcmin = 13.2




# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

# django imports

# other imports
import os
from pathlib import Path
import re
import astrometry
import numpy as np
import matplotlib as mplt
import matplotlib.patheffects
import matplotlib.pyplot as plt
import astropy.units as u
from photutils.aperture import CircularAperture
from astropy.wcs.utils import fit_wcs_from_points
from astropy.time import Time
import itertools
import datetime
import math
import gc
import yaml
from importlib import resources

# iop4lib imports
from iop4lib.enums import IMGTYPES, BANDS, OBSMODES, SRCTYPES, INSTRUMENTS, REDUCTIONMETHODS
from .instrument import Instrument
from iop4lib.utils import imshow_w_sources, get_angle_from_history, build_wcs_centered_on, get_simbad_sources
from iop4lib.utils.sourcedetection import get_segmentation, get_cat_sources_from_segment_map, get_bkg
from iop4lib.utils.plotting import plot_preview_astrometry
from iop4lib.utils.astrometry import BuildWCSResult
from iop4lib.telescopes import OSNT090

# logging
import logging
logger = logging.getLogger(__name__)


import typing
from typing import Union
if typing.TYPE_CHECKING:
    from iop4lib.db import RawFit, ReducedFit


class DIPOL(Instrument):

    name = "DIPOL"
    telescope = OSNT090.name

    instrument_kw_L = ["ASI Camera (1)"]

    arcsec_per_pix = 0.134
    field_width_arcmin = 9.22
    field_height_arcmin = 6.28 
    
    gain_e_adu = 1

    required_masters = ['masterbias', 'masterflat', 'masterdark']


    # pre computed pairs distances to use in the astrometric calibrations
    # obtained from calibrated photometry fields
    
    disp_sign_mean, disp_sign_std = np.array([-2.09032765e+02,  1.65384209e-02]), np.array([4.13289109, 0.66159702])
    disp_mean, disp_std = np.abs(disp_sign_mean), disp_sign_std
    disp_std = np.array([15, 5])

    # pre computed image angle from ~500 full photometric fields (which dont need it as input)
    # In [19]: np.quantile(angle_L,[(1-0.9973)/2,(1-0.9545)/2,(1-0.6827)/2])
    # Out[19]: array([176.85992732, 177.06814576, 177.34785346])

    # In [20]: np.quantile(angle_L,[1-(1-0.9973)/2,1-(1-0.9545)/2,1-(1-0.6827)/2])
    # Out[20]: array([181.73222701, 181.51777191, 181.36541356])

    # In [21]: np.mean(angle_L), np.median(angle_L), np.std(angle_L)
    # Out[21]: (179.08971366048235, 177.6282921156412, 1.9341471640122656)

    default_sky_angle = 177.6
    default_sky_angle_std = 2.3

    @classmethod
    def classify_juliandate_rawfit(cls, rawfit: 'RawFit'):
        """
        DIPOL files have JD keyword
        """
        import astropy.io.fits as fits
        jd = fits.getheader(rawfit.filepath, ext=0)["JD"]
        rawfit.juliandate = jd


    @classmethod
    def classify_imgtype_rawfit(cls, rawfit: 'RawFit'):
        """
        DIPOL files have IMAGETYP keyword: Light Frame (it can b, Bias Frame

        """
        from iop4lib.db.rawfit import RawFit
        import astropy.io.fits as fits

        with fits.open(rawfit.filepath) as hdul:
            if hdul[0].header['IMAGETYP'] == 'Bias Frame':
                rawfit.imgtype = IMGTYPES.BIAS
            elif hdul[0].header['IMAGETYP'] == 'Dark Frame':
                rawfit.imgtype = IMGTYPES.DARK
            elif hdul[0].header['IMAGETYP'] == 'Flat Field':
                rawfit.imgtype = IMGTYPES.FLAT
            elif hdul[0].header['IMAGETYP'] == 'Light Frame':
                if 'skyflat' in rawfit.header['OBJECT'].lower():
                    rawfit.imgtype = IMGTYPES.FLAT
                else:
                    rawfit.imgtype = IMGTYPES.LIGHT
            else:
                logger.error(f"Unknown image type for {rawfit.fileloc}.")
                rawfit.imgtype = IMGTYPES.ERROR
                raise ValueError

    @classmethod
    def classify_band_rawfit(cls, rawfit: 'RawFit'):
        """
            .. warning: 
                Red is in a differnt photometric system.
        """

        from iop4lib.db.rawfit import RawFit

        if 'FILTER' not in rawfit.header:
            if rawfit.imgtype == IMGTYPES.BIAS or rawfit.imgtype == IMGTYPES.DARK:
                rawfit.band = BANDS.NONE
            else:
                rawfit.band = BANDS.ERROR
                raise ValueError(f"Missing FILTER keyword for {rawfit.fileloc} which is not a bias or dark (it is a {rawfit.imgtype}).")
        elif rawfit.header['FILTER'] == "Red":  # TODO: they are not exacty red, they are in a different photometric system. Conversion must be implemented.
            rawfit.band = BANDS.R
        elif rawfit.header['FILTER'] == "Green":
            rawfit.band = BANDS.V
        elif rawfit.header['FILTER'] == "Blue":
            rawfit.band = BANDS.B
        else:
            rawfit.band = BANDS.ERROR
            raise ValueError(f"Unknown FILTER keyword for {rawfit.fileloc}: {rawfit.header['FILTER']}.")
    

    @classmethod
    def classify_obsmode_rawfit(cls, rawfit: 'RawFit'):
        """
        As of 2023-10-28, DIPOL polarimetry files have NOTES keyword with the angle like 'xxxx deg',
        photometry files have empty NOTES keyword.
        """

        if 'NOTES' in rawfit.header and not 'deg' in rawfit.header['NOTES']:
            rawfit.obsmode = OBSMODES.PHOTOMETRY
        else:
            rawfit.obsmode = OBSMODES.POLARIMETRY
            try:
                rawfit.rotangle = float(rawfit.header['NOTES'].split(' ')[0])
            except Exception as e:
                logger.error(f"Error parsing NOTES keyword for {rawfit.fileloc} as a float: {e}.")


    @classmethod
    def request_master(cls, rawfit, model, other_epochs=False):
        r""" Overriden Instrument associate_masters.
        
        DIPOL POLARIMETRY files are a cut of the full field, so when associating master calibration files, it needs to search a different size of images.
        For DIPOL PHOTOMETRY files, everything is the same as in the parent class.

        The full field images are 4144x2822, polarimetry images are 1100x900, the cut 
        position is saved in the images as 
            XORGSUBF 	0
            YORGSUBF 	0
        """

        if rawfit.obsmode == OBSMODES.PHOTOMETRY:
            return super().request_master(rawfit, model, other_epochs=other_epochs)
        
        # POLARIMETRY (see docstring)
        # everything should be the same as in the parent class except for the line changing args["imgsize"]

        from iop4lib.db import RawFit

        rf_vals = RawFit.objects.filter(id=rawfit.id).values().get()
        args = {k:rf_vals[k] for k in rf_vals if k in model.margs_kwL}
        
        args.pop("exptime", None) # exptime might be a building keywords (for flats and darks), but masters with different exptime can be applied
        args["epoch"] = rawfit.epoch # from .values() we only get epoch__id 

        args["imgsize"] = "4144x2822" # search for full field calibration frames

        master = model.objects.filter(**args, flags__hasnot=model.FLAGS.IGNORE).first()
        
        if master is None and other_epochs == True:
            args.pop("epoch")

            master_other_epochs = np.array(model.objects.filter(**args, flags__hasnot=model.FLAGS.IGNORE).all())

            if len(master_other_epochs) == 0:
                logger.debug(f"No {model._meta.verbose_name} for {args} in DB, None will be returned.")
                return None
            
            master_other_epochs_jyear = np.array([md.epoch.jyear for md in master_other_epochs])
            master = master_other_epochs[np.argsort(np.abs(master_other_epochs_jyear - rawfit.epoch.jyear))[0]]
            
            if (master.epoch.jyear - rawfit.epoch.jyear) > 7/365:
                logger.warning(f"{model._meta.verbose_name} from epoch {master.epoch} is more than 1 week away from epoch {rawfit.epoch}.")
                        
        return master

    @classmethod
    def apply_masters(cls, reducedfit):
        """ Overriden for DIPOL (see DIPOL.request_master).
        
        The cut position is saved in the raw fit header as:
            XORGSUBF 	1500
            YORGSUBF 	1000
        """

        x_start = reducedfit.rawfit.header['XORGSUBF']
        y_start = reducedfit.rawfit.header['YORGSUBF']

        x_end = x_start + reducedfit.rawfit.header['NAXIS1']
        y_end = y_start + reducedfit.rawfit.header['NAXIS2']

        idx = np.s_[y_start:y_end, x_start:x_end]

        import astropy.io.fits as fits

        logger.debug(f"{reducedfit}: applying masters")

        rf_data = fits.getdata(reducedfit.rawfit.filepath)
        mb_data = fits.getdata(reducedfit.masterbias.filepath)[idx]

        if reducedfit.obsmode == OBSMODES.PHOTOMETRY:
            mf_data = fits.getdata(reducedfit.masterflat.filepath)[idx]
        else: # no flat fielding for polarimetry
            mf_data = 1.0

        if reducedfit.masterdark is not None:
            md_dark = fits.getdata(reducedfit.masterdark.filepath)[idx]
        else :
            logger.warning(f"{reducedfit}: no masterdark found, assuming dark current = 0, is this a CCD camera and it's cold?")
            md_dark = 0.0

        data_new = (rf_data - mb_data - md_dark*reducedfit.rawfit.exptime) / (mf_data)

        header_new = fits.Header()

        if not os.path.exists(os.path.dirname(reducedfit.filepath)):
            logger.debug(f"{reducedfit}: creating directory {os.path.dirname(reducedfit.filepath)}")
            os.makedirs(os.path.dirname(reducedfit.filepath))
        
        fits.writeto(reducedfit.filepath, data_new, header=header_new, overwrite=True)


    @classmethod
    def get_header_hintobject(self, rawfit):
        r""" Overriden for DIPOL, which are using the convention for the other_names field. 
        
        The regex used has been obtained from the notebook checking all keywords.
        """
        

        from iop4lib.db import AstroSource

        catalog = AstroSource.objects.exclude(is_calibrator=True).all()

        #pattern = re.compile(r"^([a-zA-Z0-9]{4,}|[a-zA-Z0-9]{1,3}(_[a-zA-Z0-9]+)?)(?=_|$)")
        pattern = re.compile(r"^([a-zA-Z0-9]{1,3}_[a-zA-Z0-9]+|[a-zA-Z0-9]{4,})(?=_|$)")
        
        obj_kw = rawfit.header['OBJECT']
        
        match = pattern.match(obj_kw)

        def get_invariable_str(s):
            return s.replace(' ', '').replace('-','').replace('+','').replace('_','').upper()

        if match:
            
            search_str = match.group(0)
            
            for source in catalog:
                if not source.other_names_list:
                    continue
                if any([get_invariable_str(search_str) in get_invariable_str(other_name) for other_name in source.other_names_list]):
                    return source
 
            for source in catalog:
                if get_invariable_str(search_str) in get_invariable_str(source.name):
                    return source
                
        return None
      
    
    @classmethod
    def get_header_hintcoord(cls, rawfit):
        """ Overriden for DIPOL

        As of 2023-10-23, DIPOL does not inclide RA and DEC in the header, RA and DEC will be derived from the object name.
        """     
        
        return rawfit.header_hintobject.coord            
            
        
    @classmethod
    def get_astrometry_size_hint(cls, rawfit: 'RawFit'):
        """ Implement Instrument.get_astrometry_size_hint for DIPOL.

            For DIPOL in OSN-T090, according to preliminary investigation of OSN crew is:
                Las posiciones que he tomado y el ángulo de rotación en cada caso son estos:
                Dec= -10º HA=+3h  rotación=-177.3º
                Zenit rotación=-177.3º
                Dec=+60º HA=-6h rotación=-177.7º
                Dec=+70º HA=+5h rotación=-177.2º

                El campo es de 9.22 x 6.28 arcmin y el tamaño de pixel de 0.134"/pix

                El ángulo de la imagen cambia muy poco entre las posiciones muy separadas del telescopio, y es de 177.5º ± 0.3º
                Así que como mucho se produce un error de ± 0.3º en las imágenes, y el punto cero es de 2.5º.
        """

        return astrometry.SizeHint(lower_arcsec_per_pixel=0.95*cls.arcsec_per_pix, upper_arcsec_per_pixel=1.05*cls.arcsec_per_pix)
    
    @classmethod
    def get_astrometry_position_hint(cls, rawfit: 'RawFit', allsky=False, n_field_width=1.5, hintsep=None):
        """ Get the position hint from the FITS header as an astrometry.PositionHint.
        
        Parameters
        ----------
        allsky: bool, optional
            If True, the hint will cover the whole sky, and n_field_width and hintsep will be ignored.
        n_field_width: float, optional
            The search radius in units of field width. Default is 1.5.
        hintsep: Quantity, optional
            The search radius in units of degrees.
        """        

        hintcoord = cls.get_header_hintcoord(rawfit)

        if rawfit.header["XBINNING"] != 2:
            logger.error(f"Cannot compute astrometry for {rawfit} because of the binning: {rawfit.header['XBINNING']}.")
            return None
        
        if allsky:
            hintsep = 180.0 * u.deg
        else:
            if hintsep is None:
                hintsep = (n_field_width * cls.field_width_arcmin*u.Unit("arcmin"))

        return astrometry.PositionHint(ra_deg=hintcoord.ra.deg, dec_deg=hintcoord.dec.deg, radius_deg=hintsep.to_value(u.deg))
    
    @classmethod
    def has_pairs(cls, fit_instance: Union['ReducedFit', 'RawFit']) -> bool:
        """ DIPOL ALWAYS HAS PAIRS """
        return True





    @classmethod
    def _estimate_positions_from_segments(cls, redf=None, data=None, fwhm=None, npixels=64, n_seg_threshold=3.0, centering=2/3):

        if fwhm is None:
            if redf is not None and redf.header_hintobject.srctype == SRCTYPES.STAR and redf.exptime <= 5:
                # the star might be too bright, we need to smooth the image not to get too fake sources
                fwhm = 80.0
            else:
                fwhm = 1.0

        # get the sources positions

        if data is None:
            data = redf.mdata
 
        height, width = data.shape

        bkg = get_bkg(data, filter_size=5, box_size=width//10)
        imgdata_bkg_substracted = data - bkg.background
        seg_threshold = n_seg_threshold * bkg.background_rms
        
        segment_map, convolved_data = get_segmentation(imgdata_bkg_substracted, fwhm=fwhm, npixels=npixels, threshold=seg_threshold)
        if segment_map is None:
            return list()
        else:
            seg_cat, positions, tb = get_cat_sources_from_segment_map(segment_map, imgdata_bkg_substracted, convolved_data)
        
        if centering:
            # select only the sources in the center
            cx, cy = width//2, height//2
            idx = np.abs(positions[:,0]-cx) < centering * width / 2
            idx = idx & (np.abs(positions[:,1]-cy) < centering * height / 2)
            positions = positions[idx]

        return positions

    @classmethod
    def build_wcs(cls, reducedfit: 'ReducedFit', summary_kwargs : dict = None, method=None):
        """ Overriden Instrument build_wcs.
        
        While for PHOTOMETRY observations, DIPOL has a wide field which can be astrometrically calibrated, 
        POLARIMETRY files are small with only the source field ordinary and extraordianty images in the center (to save up space).
        In some ocassions, there might be some close source also in the field.

        Therefore, to calibrate polarimetry files, we just give it a WCS centered on the source.

        For PHOTOMETRY files, we use the parent class method, but we set some custom shotgun_params_kwargs to account
        for the low flux and big size of the images.

        """

        if summary_kwargs is None:
            summary_kwargs = {'build_summary_images':True, 'with_simbad':True}

        if method is not None:
            logger.warning(f"Calling {method} for {reducedfit}.")
            return method(reducedfit, summary_kwargs=summary_kwargs)

        from iop4lib.db import ReducedFit, AstroSource

        target_src = reducedfit.header_hintobject

        if reducedfit.obsmode == OBSMODES.PHOTOMETRY:
            return super().build_wcs(reducedfit, shotgun_params_kwargs=cls._build_shotgun_params(reducedfit), summary_kwargs=summary_kwargs)
        elif reducedfit.obsmode == OBSMODES.POLARIMETRY:

            # Gather some info to perform a good decision on which methods to use

            n_estimate = len(cls._estimate_positions_from_segments(redf=reducedfit, n_seg_threshold=1.3, npixels=64, centering=None))
            n_estimate_centered = len(cls._estimate_positions_from_segments(redf=reducedfit, n_seg_threshold=1.3, npixels=64, centering=2/3))
            redf_phot = ReducedFit.objects.filter(instrument=reducedfit.instrument,
                                                  sources_in_field__in=[reducedfit.header_hintobject], 
                                                  obsmode=OBSMODES.PHOTOMETRY, 
                                                  flags__has=ReducedFit.FLAGS.BUILT_REDUCED).first()
            try:
                n_expected_simbad_sources = len(get_simbad_sources(reducedfit.header_hintobject.coord, radius=(reducedfit.width*cls.arcsec_per_pix*u.arcsec)))
            except Exception as e:
                logger.error(f"Error getting simbad sources for {reducedfit}: {e}")
                n_expected_simbad_sources = None
            n_expected_calibrators = AstroSource.objects.filter(calibrates__in=[reducedfit.header_hintobject]).count()

            # log the variables above

            logger.debug(f"{target_src.srctype=}")
            logger.debug(f"{n_estimate=}")
            logger.debug(f"{n_estimate_centered=}")
            logger.debug(f"{redf_phot=}")
            logger.debug(f"{n_expected_simbad_sources=}")
            logger.debug(f"{n_expected_calibrators=}")
            
            with open(resources.files("iop4lib.instruments") / "dipol_astrometry.yaml") as f:
                dipol_astrometry = yaml.safe_load(f)

            def apply_comparison(val1, val2, op):
                if op == "eq":
                    return val1 == val2
                elif op == "ne":
                    return val1 != val2
                elif op == "gt":
                    return val1 > val2
                elif op == "gte":
                    return val1 >= val2
                elif op == "lt":
                    return val1 < val2
                elif op == "lte":
                    return val1 <= val2
                elif op is None or op == "":
                    return val1 == val2

            def check_conditions(branch, context):
                if 'conds' in branch and len(branch['conds']) > 0:
                    keys, ops, vals = zip(*[(*k.split('__'),v) if '__' in k else (k,None,v) for k,v in branch['conds'].items()])
                    satisfies_conds = all([apply_comparison(context[k], v, op) for k,op,v in zip(keys, ops, vals)])
                else:
                    satisfies_conds = True
                return satisfies_conds

            def find_attempts(rules, context):
                for rule in rules:
                    if check_conditions(rule, context):
                        if 'attempts' in rule:
                            # get the attempts as a list, fetching the definitions if necessary
                            attempts = [dipol_astrometry['attempt_defs'][attempt] if isinstance(attempt, str) else attempt for attempt in rule['attempts']]
                            # filter the attempts by the conditions
                            return list(filter(lambda attempt: check_conditions(attempt, context), attempts))
                        else:
                            return find_attempts(rule['rules'], context)
                return None
        
            attempts = find_attempts(dipol_astrometry['rules'], context = {
                'n_estimate': n_estimate,
                'n_estimate_centered': n_estimate_centered,
                'n_expected_simbad_sources': n_expected_simbad_sources,
                'n_expected_calibrators': n_expected_calibrators,
                'exptime': reducedfit.exptime,
                'srctype': target_src.srctype,
                'srcname': target_src.name,
                'redf_phot': redf_phot,
            })

            
            for attempt in attempts:
                logger.debug(f"Trying attempt: {attempt} for {reducedfit}.")

                m = getattr(cls, attempt['method'])
                args = attempt['args']

                args_keys = list(args.keys())
                args_values_list = list(itertools.product(*args.values()))
                args_dict_list = [dict(zip(args_keys, args_vals)) for args_vals in args_values_list]

                build_wcs = False

                for args_dict in args_dict_list:
                    logger.info(f"Attempt: {attempt}: {m.__name__} with {args_dict} for {reducedfit}.")
                    if (build_wcs := m(reducedfit, summary_kwargs=summary_kwargs, **args_dict)):
                        break
                
                build_wcs.info["attempt"] = attempt
                build_wcs.info["m.__name__"] = m.__name__
                build_wcs.info["args"] = args_dict
                build_wcs.info["n_estimate"] = n_estimate
                build_wcs.info["n_estimate_centered"] = n_estimate_centered
                build_wcs.info["n_expected_simbad_sources"] = n_expected_simbad_sources
                build_wcs.info["n_expected_calibrators"] = n_expected_calibrators

                if build_wcs:
                    break

            return build_wcs
        
        else:
            logger.error(f"Unknown obsmode {reducedfit.obsmode} for {reducedfit}.")
            raise ValueError
            
        
    @classmethod
    def _build_shotgun_params(cls, redf: 'ReducedFit'):
        shotgun_params_kwargs = dict()

        shotgun_params_kwargs["keep_n_seg"] = [300]
        shotgun_params_kwargs["border_margin_px"] = [20]
        shotgun_params_kwargs["output_logodds_threshold"] = [14]
        shotgun_params_kwargs["n_rms_seg"] = [1.5, 1.2, 1.0]
        shotgun_params_kwargs["bkg_filter_size"] = [11] 
        shotgun_params_kwargs["bkg_box_size"] = [32]
        shotgun_params_kwargs["seg_fwhm"] = [1.0]
        shotgun_params_kwargs["npixels"] = [32, 8]
        shotgun_params_kwargs["seg_kernel_size"] = [None]
        shotgun_params_kwargs["allsky"] = [False]

        shotgun_params_kwargs["d_eps"] = [4.0]
        shotgun_params_kwargs["dx_eps"] = [4.0]
        shotgun_params_kwargs["dy_eps"] = [2.0]
        shotgun_params_kwargs["dx_min"] = [150]
        shotgun_params_kwargs["dx_max"] = [300]
        shotgun_params_kwargs["dy_min"] = [0]
        shotgun_params_kwargs["dy_max"] = [50]
        shotgun_params_kwargs["d_min"] = [150]
        shotgun_params_kwargs["d_max"] = [250]
        shotgun_params_kwargs["bins"] = [400]
        shotgun_params_kwargs["hist_range"] = [(0,500)]

        shotgun_params_kwargs["position_hint"] = [redf.get_astrometry_position_hint(allsky=False)]
        shotgun_params_kwargs["size_hint"] = [redf.get_astrometry_size_hint()]

        return shotgun_params_kwargs


    @classmethod
    def _build_wcs_for_polarimetry_images_photo_quads(cls, redf: 'ReducedFit', summary_kwargs : dict = None, n_seg_threshold=1.5, npixels=32, min_quad_distance=4.0, fwhm=None, centering=None, max_quad_t=1000, min_quad_area=0.03):

        if summary_kwargs is None:
            summary_kwargs = {'build_summary_images':True, 'with_simbad':True}

        from iop4lib.db import ReducedFit

        if (target_src := redf.header_hintobject) is None:
            raise Exception("No target source found in header, cannot build WCS.")

        # the polarimetry field

        redf_pol = redf

        # select an already solved photometry field

        redf_phot = ReducedFit.objects.filter(instrument=redf_pol.instrument, 
                                              sources_in_field__in=[target_src], 
                                              obsmode=OBSMODES.PHOTOMETRY, 
                                              flags__has=ReducedFit.FLAGS.BUILT_REDUCED).first()
        
        if redf_phot is None:
            logger.error(f"No astro-calibrated photometry field found for {redf_pol}.")
            return BuildWCSResult(success=False)

        logger.debug(f"Invoked with {n_seg_threshold=}, {npixels=}")

        # get the subframe of the photometry field that corresponds to this polarimetry field, (approx)
        x_start = redf_pol.rawfit.header['XORGSUBF']
        y_start = redf_pol.rawfit.header['YORGSUBF']

        x_end = x_start + redf_pol.rawfit.header['NAXIS1']
        y_end = y_start + redf_pol.rawfit.header['NAXIS2']

        idx = np.s_[y_start:y_end, x_start:x_end]

        photdata_subframe = redf_phot.mdata[idx] # if we use the hash_ish_old, which is not invariant under fliiping, we need to flip the image in y (redf_phot.mdata[idx][::-1,:])

        # find 10 brightest sources in each field

        sets_L = list()

        for redf, data in zip([redf_pol, redf_phot], [redf_pol.mdata, photdata_subframe]):

            positions = cls._estimate_positions_from_segments(redf=redf, data=data, n_seg_threshold=n_seg_threshold, npixels=npixels, centering=centering, fwhm=fwhm)
            positions = positions[:10]

            sets_L.append(positions)
        
        logger.debug(f"Using {len(sets_L[0])} sources in polarimetry field and {len(sets_L[1])} in photometry field.")

        if summary_kwargs['build_summary_images']:
            logger.debug(f"Building summary image for astrometry detected sources.")

            fig = mplt.figure.Figure(figsize=(12,6), dpi=iop4conf.mplt_default_dpi)
            axs = fig.subplots(nrows=1, ncols=2)

            for ax, data, positions in zip(axs, [redf_pol.mdata, photdata_subframe], sets_L):
                imshow_w_sources(data, pos1=positions, ax=ax)
                candidates_aps = CircularAperture(positions[:2], r=10.0)
                candidates_aps.plot(ax, color="b")
                for i, (x,y) in enumerate(positions):
                    ax.text(x, y, f"{i}", color="orange", fontdict={"size":14, "weight":"bold"})#, verticalalignment="center", horizontalalignment="center") 
                ax.plot([data.shape[1]//2], [data.shape[0]//2], '+', color='y', markersize=10)
                
            axs[0].set_title("Polarimetry field")
            axs[1].set_title("Photometry field")
            fig.savefig(Path(redf_pol.filedpropdir) / "astrometry_detected_sources.png", bbox_inches="tight")
            fig.clf()

        # Build the quads for each field
        quads_1 = np.array(list(itertools.combinations(sets_L[0], 4)))
        quads_2 = np.array(list(itertools.combinations(sets_L[1], 4)))

        # remove quads of points that have an area less than min_quad_area * area of the image

        def PolyArea(x,y):
            # order points clockwise
            idx = np.argsort(np.arctan2(y-y.mean(), x-x.mean()))
            x, y = x[idx], y[idx]
            return 0.5*np.abs(np.dot(x,np.roll(y,1))-np.dot(y,np.roll(x,1)))
        
        quads_1 = np.array([quad for quad in quads_1 if PolyArea(quad[:,0], quad[:,1]) > min_quad_area*(redf_pol.width*redf_pol.height)])
        quads_2 = np.array([quad for quad in quads_2 if PolyArea(quad[:,0], quad[:,1]) > min_quad_area*(redf_pol.width*redf_pol.height)])

        if len(quads_1) == 0 or len(quads_2) == 0:
            logger.error(f"No quads found in {redf_pol} and {redf_phot}, returning success = False.")
            return BuildWCSResult(success=False)

        # Compute the hashes of the quads
      
        from iop4lib.utils.quadmatching import hash_ish, distance, order, qorder_ish
        hash_func, qorder = hash_ish, qorder_ish

        hashes_1 = np.array([hash_func(quad) for quad in quads_1])
        hashes_2 = np.array([hash_func(quad) for quad in quads_2])

        # Get the distance (in the hash space) between all quads

        all_indices = np.array(list(itertools.product(range(len(quads_1)),range(len(quads_2)))))
        all_distances = np.array([distance(hashes_1[i], hashes_2[j]) for i,j in all_indices])

        idx = np.argsort(all_distances)
        all_indices = all_indices[idx]
        all_distances = all_distances[idx]

        # select some nice indices (the best 5 that have less than min_quad_distance of error per quad)
 
        idx_selected = np.where(all_distances < min_quad_distance)[0] 
        indices_selected = all_indices[idx_selected]
        distances_selected = all_distances[idx_selected]
        
        if len(idx_selected) == 0:
            logger.error(f"No quads with distance < {min_quad_distance}, minimum at {min(all_distances)=} returning success = False.")
            return BuildWCSResult(success=False, wcslist=None, info={'redf_phot__pk':redf_phot.pk, 'redf_phot__fileloc':redf_phot.fileloc}) 
        else:
            idx_selected = np.argsort(distances_selected)[:5]
            indices_selected = all_indices[idx_selected]
            distances_selected = all_distances[idx_selected]


        # save the flipped status of both images

        is_redf_pol_flipped = 'FLIPSTAT' in redf_pol.rawfit.header and redf_pol.rawfit.header['FLIPSTAT'] == "Flip"
        is_redf_phot_flipped = 'FLIPSTAT' in redf_phot.rawfit.header and redf_phot.rawfit.header['FLIPSTAT'] == "Flip"

        # Get the appropiate transformation depending on whether both images are flipped or not

        from iop4lib.utils.quadmatching import find_best_transformation, distance_to_y_flip, distance_to_identity
        
        if is_redf_pol_flipped != is_redf_phot_flipped:
            dist_func =  distance_to_y_flip
        else:
            dist_func =  distance_to_identity

        # get linear transforms
        logger.debug(f"Selected {len(indices_selected)} quads with distance < {min_quad_distance}. I will get the one with less deviation from the median linear transform.")

        R_L, t_L, perm_L = zip(*[find_best_transformation(quads_1[i], quads_2[j], dist_func) for i,j in indices_selected])
        logger.debug(f"{t_L=}")
        

        logger.debug(f"Filtering out big translations (<{max_quad_t} px)")

        _indices_selected = indices_selected[np.array([np.linalg.norm(t) < max_quad_t for t in t_L])]

        logger.debug(f"Filtered to {len(_indices_selected)} quads with distance < {min_quad_distance} and translation < {max_quad_t} px.")

        if len(_indices_selected) == 0:
            logger.error(f"No quads with distance < {min_quad_distance} and translation < {max_quad_t} px, building summary image of the 3 best quads and returning success = False.")

            colors = [color for color in mplt.rcParams["axes.prop_cycle"].by_key()["color"]]

            fig = mplt.figure.Figure(figsize=(12,6), dpi=iop4conf.mplt_default_dpi)
            axs = fig.subplots(nrows=1, ncols=2)

            for (i, j), color in list(zip(indices_selected, colors))[:3]: 

                tij = find_best_transformation(quads_1[i], quads_2[j], dist_func)[1]

                for ax, data, quad, positions in zip(axs, [redf_pol.mdata, photdata_subframe], [quads_1[i], quads_2[j]], sets_L):
                    imshow_w_sources(data, pos1=positions, ax=ax)
                    x, y = np.array(order(quad)).T
                    ax.fill(x, y, edgecolor='k', fill=True, facecolor=mplt.colors.to_rgba(color, alpha=0.2))
                    for pi, p in enumerate(np.array((qorder(quad)))):
                        xp = p[0]
                        yp = p[1]
                        ax.text(xp, yp, f"{pi}", fontsize=16, color=color, path_effects=[mplt.patheffects.Stroke(linewidth=1, foreground='black'), mplt.patheffects.Normal()])

                fig.suptitle(f"dist({i},{j})={distance(hash_func(quads_1[i]),hash_func(quads_2[j])):.3f}, norm(t) = {np.linalg.norm(tij):.0f} px", y=0.83)

            axs[0].set_title("Polarimetry")
            axs[1].set_title("Photometry")
            
            fig.savefig(Path(redf_pol.filedpropdir) / "astrometry_matched_quads.png", bbox_inches="tight")
            fig.clf()

            return BuildWCSResult(success=False, wcslist=None, info={'redf_phot__pk':redf_phot.pk, 'redf_phot__fileloc':redf_phot.fileloc})
        else:
            indices_selected = _indices_selected
        
        R_L, t_L, perm_L = zip(*[find_best_transformation(quads_1[i], quads_2[j], dist_func) for i,j in indices_selected])


        # get the closest one to the t_L mean
        median_t = np.median(t_L, axis=0)

        logger.debug(f"{median_t=}")

        delta_t = np.array([np.linalg.norm(t - median_t) for t in t_L])
        indices_selected = indices_selected[np.argsort(delta_t)]
        best_i, best_j = indices_selected[0]
        
        logger.debug(f"Selected the quads [{best_i},{best_j}]")

        t = t_L[np.argmin(delta_t)]
        R = R_L[np.argmin(delta_t)]
        logger.debug(f"t = {t}, R = {R}")
        logger.debug(f"det R = {np.linalg.det(R)}")

        # build_summary_images, replace indices_selected by all_indices if no R,t filtering was done.
        if summary_kwargs['build_summary_images']:
            logger.debug(f"Building summary image for quad matching.")

            colors = [color for color in mplt.rcParams["axes.prop_cycle"].by_key()["color"]]

            fig = mplt.figure.Figure(figsize=(12,6), dpi=iop4conf.mplt_default_dpi)
            axs = fig.subplots(nrows=1, ncols=2)

            for (i, j), color in list(zip(indices_selected, colors))[:1]: 
                
                tij = find_best_transformation(quads_1[i], quads_2[j], dist_func)[1]

                for ax, data, quad, positions in zip(axs, [redf_pol.mdata, photdata_subframe], [quads_1[i], quads_2[j]], sets_L):
                    imshow_w_sources(data, pos1=positions, ax=ax)
                    x, y = np.array(order(quad)).T
                    ax.fill(x, y, edgecolor='k', fill=True, facecolor=mplt.colors.to_rgba(color, alpha=0.2))
                    for pi, p in enumerate(np.array((qorder(quad)))):
                        xp = p[0]
                        yp = p[1]
                        ax.text(xp, yp, f"{pi}", fontsize=16, color=color, path_effects=[mplt.patheffects.Stroke(linewidth=1, foreground='black'), mplt.patheffects.Normal()])

                fig.suptitle(f"dist({i},{j})={distance(hash_func(quads_1[i]),hash_func(quads_2[j])):.3f}, norm(t) = {np.linalg.norm(tij):.0f} px", y=0.83)

            axs[0].set_title("Polarimetry")
            axs[1].set_title("Photometry")
            
            fig.savefig(Path(redf_pol.filedpropdir) / "astrometry_matched_quads.png", bbox_inches="tight")
            fig.clf()

        # Build the WCS

        # give an unique ordering to the quads

        quads_1 = [qorder_ish(quad) for quad in quads_1]
        quads_2 = [qorder_ish(quad) for quad in quads_2]

        # get the pre wcs with the target in the center of the image (if the image is flipped, the angle is negative)

        angle_mean, angle_std = get_angle_from_history(redf_pol, target_src)
        if is_redf_pol_flipped:
            angle = - angle_mean
        else:
            angle = angle_mean

        logger.debug(f"Using {angle=} for pre wcs.")

        # fit a wcs centered on the target source

        pre_wcs = build_wcs_centered_on((redf_pol.width//2,redf_pol.height//2), redf=redf_phot, angle=angle)

        # Get the pixel position of the quad points in the (small) polarimetry field
        pix_array_1 = np.array(list(zip(*[(x,y) for x,y in quads_1[best_i]])))

        # Get pixel positions of the quad points in the (full) photometry field
        #pix_array_2 = np.array(list(zip(*[(x+x_start,y+y_start) for x,y in quads_2[best_j]])))
        # instead of quads_2[best_j], transform quads_1[best_i] with the linear transformation, to avoid
        # incorrect permutations
        pix_array_2 = np.array(list(zip(*[(x+x_start,y+y_start) for x,y in (np.dot(R, np.array(quads_1[best_i]).T).T + t)])))

        # fit the WCS so the pixel arrays in 1 correspond to the ra/dec of the pixel array in 2
        wcs1 = fit_wcs_from_points(pix_array_1,  redf_phot.wcs1.pixel_to_world(*pix_array_2), projection=pre_wcs)
        wcs2 = fit_wcs_from_points(pix_array_1,  redf_phot.wcs2.pixel_to_world(*pix_array_2), projection=pre_wcs)

        wcslist = [wcs1, wcs2]

        if summary_kwargs['build_summary_images']:
            logger.debug(f"Building summary image for astrometry.")
            fig = mplt.figure.Figure(figsize=(6,6), dpi=iop4conf.mplt_default_dpi)
            ax = fig.subplots(nrows=1, ncols=1, subplot_kw={'projection': wcslist[0]})
            plot_preview_astrometry(redf_pol, with_simbad=True, has_pairs=True, wcs1=wcslist[0], wcs2=wcslist[1], ax=ax, fig=fig) 
            fig.savefig(Path(redf_pol.filedpropdir) / "astrometry_summary.png", bbox_inches="tight")
            fig.clear()            


        result = BuildWCSResult(success=True, wcslist=wcslist, info={'redf_phot__pk':redf_phot.pk, 'redf_phot__fileloc':redf_phot.fileloc, 'n_seg_threshold':n_seg_threshold, 'npixels':npixels})

        return result    



    @classmethod
    def _build_wcs_for_polarimetry_from_target_O_and_E(cls, redf: 'ReducedFit', summary_kwargs : dict = None, n_seg_threshold=3.0, npixels=64, fwhm=None, centering=2/3) -> BuildWCSResult:
        r""" Deprecated. Build WCS for DIPOL polarimetry images by matching the found sources positions with the catalog.

        .. warning::
            This method is deprecated and will be removed in the future. It is kept here for reference. Do not use, unless you know what you are doing.

            When there are some DB sources in the field it probably works, but when there are not, it might give a result that is centered on the wrong source!
        
            It is safer to use the method _build_wcs_for_polarimetry_images_photo_quads, which uses the photometry field and quad matching to build the WCS.
            
        """

        if summary_kwargs is None:
            summary_kwargs = {'build_summary_images':True, 'with_simbad':True}

        # disp_allowed_err = 1.5*cls.disp_std
        disp_allowed_err =  np.array([30,30]) # most times should be much smaller (1.5*std)
        # but in bad cases, this is ~1 sigma of the gaussians

        from iop4lib.db import AstroSource
        from iop4lib.utils.sourcepairing import get_best_pairs
        from iop4lib.utils.quadmatching import distance

        logger.debug(f"{redf}: building WCS for DIPOL polarimetry images from target_O and target_E with {npixels=}, {n_seg_threshold=}.")
        
        # definitions

        target_src = redf.header_hintobject

        if target_src is None:
            raise Exception("No target source found in header, cannot build WCS.")
    
        data = redf.mdata
        
        # get the sources positions

        cx, cy = redf.width//2, redf.height//2
        positions = cls._estimate_positions_from_segments(redf=redf, n_seg_threshold=n_seg_threshold, npixels=npixels, fwhm=fwhm, centering=centering)

        if len(positions) == 0:
            logger.error(f"{redf}: Found no sources in the field, cannot build WCS.")
            return BuildWCSResult(success=False)

        if summary_kwargs['build_summary_images']:
            # plot summary of detected sources
            fig = mplt.figure.Figure(figsize=(6,6), dpi=iop4conf.mplt_default_dpi)
            ax = fig.subplots(nrows=1, ncols=1)
            imshow_w_sources(data, pos1=positions, ax=ax)
            candidates_aps = CircularAperture(positions[:2], r=10.0)
            candidates_aps.plot(ax, color="b")
            for i, (x,y) in enumerate(positions[:15]):
                ax.text(x, y, f"{i}", color="orange", fontdict={"size":14, "weight":"bold"})#, verticalalignment="center", horizontalalignment="center") 
            ax.plot([cx], [cy], '+', color='y', markersize=10)
            if centering:
                ax.axhline(cy-centering*redf.height/2, xmin=0, xmax=redf.width, color='y', linestyle='--')
                ax.axhline(cy+centering*redf.height/2, xmin=0, xmax=redf.width, color='y', linestyle='--')
                ax.axvline(cx-centering*redf.width/2, ymin=0, ymax=redf.height, color='y', linestyle='--')
                ax.axvline(cx+centering*redf.width/2, ymin=0, ymax=redf.height, color='y', linestyle='--')
            fig.savefig(Path(redf.filedpropdir) / "astrometry_detected_sources.png", bbox_inches="tight")
            fig.clear()
            
        if len(positions) == 1:
            logger.error(f"{redf}: Found only one source in the field, cannot build WCS.")
            return BuildWCSResult(success=False)
        
        if len(positions) > 2:
            logger.warning(f"{redf}: {len(positions)} sources found, expected 2. Maybe after looking at pairs only, we can find the right ones.")

            pre_list1, pre_list2 = zip(*itertools.product(positions, positions))

            # log some debug info about the pairs diference and the difference with respect the expected disp_sign_mean
            for i, (pos1, pos2) in enumerate(zip(pre_list1, pre_list2)):
                dist = distance(pos1, pos2)
                disp = np.abs(np.subtract(pos1, pos2))
                diff = np.abs(np.subtract(pos1, pos2))-np.abs(cls.disp_sign_mean)
                with np.printoptions(precision=1, suppress=True):
                    logger.debug(f"{i=}, {pos1=!s}, {pos2=!s}, dist={dist:.2f}, {disp=!s}, {diff=!s}")

            list1, list2, d0_new, disp_sign_new = get_best_pairs(pre_list1, pre_list2, cls.disp_sign_mean, disp_sign_err=disp_allowed_err)

            logger.debug(f"{list1=}, {list2=}, {d0_new=}, {disp_sign_new=}")

            if len(list1) == 0:
                list1, list2, d0_new, disp_sign_new = get_best_pairs(pre_list1, pre_list2, -cls.disp_sign_mean, disp_sign_err=disp_allowed_err)
                logger.debug(f"{list1=}, {list2=}, {d0_new=}, {disp_sign_new=}")
            
            if len(list1) != 1:
                logger.error(f"We expected exactly one source, but we found {len(list1)} pairs, returning success = False.")
                return BuildWCSResult(success=False)
            else:
                positions = [list1[0], list2[0]]
            

        # Check that the sources are pairs
        if not np.isclose(np.abs(positions[0][0]-positions[1][0]), cls.disp_mean[0], atol=disp_allowed_err[0]):
            logger.error(f"These are not pairs, x mismatch detected according to hard-coded pair distance: disp x = {np.abs(positions[0][0]-positions[1][0]):.0f} px")
            return BuildWCSResult(success=False, info={'n_bright_sources':len(positions)})
        if not np.isclose(np.abs(positions[0][1]-positions[1][1]), cls.disp_mean[1], atol=disp_allowed_err[1]):
            logger.error(f"These are not pairs, y mismatch detected according to hard-coded pair distance: disp y = {np.abs(positions[0][1]-positions[1][1]):.0f} px")
            return BuildWCSResult(success=False, info={'n_bright_sources':len(positions)})

        # define the targets to be the two positions found, with the ordinary on the right
        target_O, target_E = positions

        if target_O[0] < target_E[0]:
            target_O, target_E = target_E, target_O

        # get the right angle
        angle_mean, angle_std = get_angle_from_history(redf, target_src)
        logger.debug(f"Using angle {angle_mean:.2f} +- {angle_std:.2f} deg")

        # DIPOL polarimery images seem to be flipped vertically, which results in negative angle
        if 'FLIPSTAT' in redf.rawfit.header and redf.rawfit.header['FLIPSTAT'] == "Flip":
            angle = - angle_mean
        else:
            angle = angle_mean

        # Build the WCS with the mean angle, and the right source in the center
        wcs1 = build_wcs_centered_on(target_O, target_src=target_src, redf=redf, angle=angle)
        wcs2 = build_wcs_centered_on(target_E, target_src=target_src, redf=redf, angle=angle)

        # TODO: try to improve wcs by finding other weaker sources and fitting a wcs with them
        # comparing them to a photometry field or DB or simbad

        if summary_kwargs['build_summary_images']:
            logger.debug(f"Building summary images for {redf}.")
            # plot summary of astrometry
            fig = mplt.figure.Figure(figsize=(6,6), dpi=iop4conf.mplt_default_dpi)
            ax = fig.subplots(nrows=1, ncols=1, subplot_kw={'projection': wcs1})
            plot_preview_astrometry(redf, with_simbad=True, has_pairs=True, wcs1=wcs1, wcs2=wcs2, ax=ax, fig=fig) 
            fig.savefig(Path(redf.filedpropdir) / "astrometry_summary.png", bbox_inches="tight")
            fig.clear()

        return BuildWCSResult(success=True, wcslist=[wcs1,wcs2],  info={'n_bright_sources':len(positions)})
    



    @classmethod
    def estimate_common_apertures(cls, reducedfits, reductionmethod=None, fit_boxsize=None, search_boxsize=(90,90)):
        aperpix, r_in, r_out, fit_res_dict = super().estimate_common_apertures(reducedfits, reductionmethod=reductionmethod, fit_boxsize=fit_boxsize, search_boxsize=search_boxsize, fwhm_min=5.0, fwhm_max=60)
        
        sigma = fit_res_dict['sigma']
        fwhm = fit_res_dict["mean_fwhm"]

        return 1.1*fwhm, 6*fwhm, 10*fwhm, fit_res_dict

    @classmethod
    def get_instrumental_polarization(cls, reducedfit) -> dict:
        """ Returns the instrumental polarization for to be used for a given reducedfit.

        The instrumental polarization is a dictionary with the following keys:
            - Q_inst: instrumental Q Stokes parameter (0-1)
            - dQ_inst: instrumental Q error (0-1)
            - U_inst: instrumental U Stokes parameter (0-1)
            - dU_inst: instrumental U error (0-1)
            - CPA: zero-angle (deg)
        """

        if reducedfit.juliandate <= Time("2023-09-28 12:00").jd: # limpieza de espejos
            CPA = 44.5
            dCPA = 0.05
            Q_inst = 0.05777
            dQ_inst = 0.005
            U_inst = -3.77095
            dU_inst = 0.005
        else:
            CPA = 45.1
            dCPA = 0.05
            Q_inst = -0.0138 / 100
            dQ_inst = 0.005
            U_inst = -4.0806 / 100
            dU_inst = 0.005

        return {'Q_inst':Q_inst, 'dQ_inst':dQ_inst, 'U_inst':U_inst, 'dU_inst':dU_inst, 'CPA':CPA, 'dCPA':dCPA}


    @classmethod
    def compute_relative_polarimetry(cls, polarimetry_group):
        """ Computes the relative polarimetry for a polarimetry group for DIPOL
        """
        
        from iop4lib.db.aperphotresult import AperPhotResult
        from iop4lib.db.photopolresult import PhotoPolResult
        from iop4lib.utils import get_column_values

        # Perform some checks on the group

        if not all([reducedfit.instrument == INSTRUMENTS.DIPOL for reducedfit in polarimetry_group]):
            raise Exception(f"This method is only for DIPOL images.")

        ## get the band of the group

        bands = [reducedfit.band for reducedfit in polarimetry_group]

        if len(set(bands)) == 1:
            band = bands[0]
        else: # should not happen
            raise Exception(f"Can not compute relative polarimetry for a group with different bands: {bands}")

        ## check obsmodes

        if not all([reducedfit.obsmode == OBSMODES.POLARIMETRY for reducedfit in polarimetry_group]):
            raise Exception(f"This method is only for polarimetry images.")
        
        ## check sources in the fields

        sources_in_field_qs_list = [reducedfit.sources_in_field.all() for reducedfit in polarimetry_group]
        group_sources = set.intersection(*map(set, sources_in_field_qs_list))

        if len(group_sources) == 0:
            logger.error("No common sources in field for all polarimetry groups.")
            return
        
        if group_sources != set.union(*map(set, sources_in_field_qs_list)):
            logger.warning("Sources in field do not match for all polarimetry groups (ReducedFit %s): %s" % (",".join([str(redf.pk) for redf in polarimetry_group]), str(set.difference(*map(set, sources_in_field_qs_list)))))

        ## check rotation angles

        rot_angles_available = set([redf.rotangle for redf in polarimetry_group])
        rot_angles_required = {0.0, 22.5, 45.0, 67.5, 90.0, 112.5, 135.0, 157.5, 180.0, 202.5, 225.0, 247.5, 270.0, 292.5, 315.0, 337.5}

        if not rot_angles_available.issubset(rot_angles_required):
            logger.warning(f"Rotation angles missing: {rot_angles_required - rot_angles_available}")

        if len(polarimetry_group) != 16:
            raise Exception(f"Can not compute relative polarimetry for a group with {len(polarimetry_group)} reducedfits, it should be 16.")

        # 1. Compute all aperture photometries

        aperpix, r_in, r_out, fit_res_dict = cls.estimate_common_apertures(polarimetry_group, reductionmethod=REDUCTIONMETHODS.RELPHOT)
        target_fwhm = fit_res_dict['mean_fwhm']
        
        logger.debug(f"Computing aperture photometries for the {len(polarimetry_group)} reducedfits in the group with target aperpix {aperpix:.1f}.")

        for reducedfit in polarimetry_group:
            cls.compute_aperture_photometry(reducedfit, aperpix, r_in, r_out)

        # 2. Compute relative polarimetry for each source (uses the computed aperture photometries)

        logger.debug("Computing relative polarimetry.")

        photopolresult_L = list()

        for astrosource in group_sources:

            if astrosource.calibrates.count() > 0:
                continue

            logger.debug(f"Computing relative polarimetry for {astrosource}.")

            # if any angle is missing for some pair, it uses the equivalent angle of the other pair

            aperphotresults = AperPhotResult.objects.filter(reducedfit__in=polarimetry_group, astrosource=astrosource, aperpix=aperpix, flux_counts__isnull=False)

            if len(aperphotresults) == 0:
                logger.error(f"No aperphotresults found for {astrosource}")
                continue

            if len(aperphotresults) != 32:
                logger.error(f"There should be 32 aperphotresults for each astrosource in the group, there are {len(aperphotresults)} for {astrosource.name}.")
                continue

            values = get_column_values(aperphotresults, ['reducedfit__rotangle', 'flux_counts', 'flux_counts_err', 'pairs'])

            angles_L = list(sorted(set(values['reducedfit__rotangle'])))

            if len(angles_L) != 16:
                logger.warning(f"There should be 16 different angles, there are {len(angles_L)}.")

            fluxD = {}
            for pair, angle, flux, flux_err in zip(values['pairs'], values['reducedfit__rotangle'], values['flux_counts'], values['flux_counts_err']):
                if pair not in fluxD:
                    fluxD[pair] = {}
                fluxD[pair][angle] = (flux, flux_err)

            # Dipol has the ordinary to the left and extraordinary to the right, IOP4 astrocalibration atm works the other way, swap them here

            F_O = np.array([(fluxD['E'][angle][0]) for angle in angles_L])
            dF_O = np.array([(fluxD['E'][angle][1]) for angle in angles_L])

            F_E = np.array([(fluxD['O'][angle][0]) for angle in angles_L])
            dF_E = np.array([(fluxD['O'][angle][1]) for angle in angles_L])

            N = len(angles_L)

            if astrosource.name == "2200+420":
                F_O = np.roll(F_E, 2)
                dF_O = np.roll(dF_E, 2)

            F = (F_O - F_E) / (F_O + F_E)
            dF = 2 / ( F_O + F_E )**2 * np.sqrt(F_E**2 * dF_O**2 + F_O**2 * dF_E**2)

            I = (F_O + F_E)
            dI = np.sqrt(dF_O**2 + dF_E**2)

            # Compute both the uncorrected and corrected values

            Qr_uncorr = 2/N * sum([F[i] * math.cos(math.pi/2*i) for i in range(N)])
            dQr_uncorr = 2/N * math.sqrt(sum([dF[i]**2 * math.cos(math.pi/2*i)**2 for i in range(N)]))

            logger.debug(f"{Qr_uncorr=}, {dQr_uncorr=}")

            Ur_uncorr = 2/N * sum([F[i] * math.sin(math.pi/2*i) for i in range(N)])
            dUr_uncorr = 2/N * math.sqrt(sum([dF[i]**2 * math.sin(math.pi/2*i)**2 for i in range(N)]))

            intrumental_polarization = cls.get_instrumental_polarization(reducedfit=polarimetry_group[0])
            Q_inst = intrumental_polarization['Q_inst']
            dQ_inst = intrumental_polarization['dQ_inst']
            U_inst = intrumental_polarization['U_inst']
            dU_inst = intrumental_polarization['dU_inst']
            CPA = intrumental_polarization['CPA']
            dCPA = intrumental_polarization['dCPA']

            logger.debug(f"{Q_inst=}, {dQ_inst=}")
            logger.debug(f"{U_inst=}, {dU_inst=}")

            Qr = Qr_uncorr - Q_inst
            dQr = math.sqrt(dQr_uncorr**2 + dQ_inst**2)

            logger.debug(f"{Qr=}, {dQr=}")

            Ur = Ur_uncorr - U_inst
            dUr = math.sqrt(dUr_uncorr**2 + dU_inst**2)

            logger.debug(f"{Ur=}, {dUr=}")

            def _get_p_and_chi(Qr, Ur, dQr, dUr):
                # linear polarization (0 to 1)
                P = math.sqrt(Qr**2+Ur**2)
                dP = 1/P * math.sqrt((Qr*dQr)**2 + (Ur*dUr)**2)

                # polarization angle (degrees)
                chi = 0.5 * math.degrees(math.atan2(Ur, Qr))
                dchi = 0.5 * math.degrees( 1 / (Qr**2 + Ur**2) * math.sqrt((Qr*dUr)**2 + (Ur*dQr)**2) )

                return P, chi, dP, dchi
            
            P_uncorr, chi_uncorr, dP_uncorr, dchi_uncorr = _get_p_and_chi(Qr_uncorr, Ur_uncorr, dQr_uncorr, dUr_uncorr)
            P, chi, dP, dchi = _get_p_and_chi(Qr, Ur, dQr, dUr)
            chi = chi + CPA
            dchi = math.sqrt(dchi**2 + dCPA**2)

            logger.debug(f"{P=}, {chi=}, {dP=}, {dchi=}")

            # No attempt to get magnitude from polarimetry fields in dipol, they have too low exposure, and many times there are no calibrators in the subframe.

            # save the results
                    
            result = PhotoPolResult.create(reducedfits=polarimetry_group, 
                                                            astrosource=astrosource, 
                                                            reduction=REDUCTIONMETHODS.RELPOL, 
                                                            p=P, p_err=dP, chi=chi, chi_err=dchi,
                                                            _q_nocorr=Qr_uncorr, _u_nocorr=Ur_uncorr, _p_nocorr=P_uncorr, _chi_nocorr=chi_uncorr,
                                                            aperpix=aperpix)

            result.aperphotresults.set(aperphotresults, clear=True)
                        
            photopolresult_L.append(result)

        # 3. Save results
        for result in photopolresult_L:
            result.save()



from django.contrib import admin

from django.utils.html import format_html
from django.urls import path, reverse 
from django.http import HttpResponse, HttpResponseRedirect, HttpResponseNotFound
from django.shortcuts import render

from iop4api.filters import *
from iop4api.models import *
from iop4admin import views

import itertools

class AdminEpoch(admin.ModelAdmin):
    model = Epoch
    list_display = ['id', 'telescope', 'night', 'status', 'count_rawfits', 'count_bias', 'count_darks', 'count_flats', 'count_light', 'count_reduced', 'count_calibrated', 'summary_rawfits_status', 'details']
    search_fields = ['id', 'telescope', 'night']
    readonly_fields = [field.name for field in Epoch._meta.fields] 
    ordering = ['-night','-telescope']
    list_filter = ['telescope', 'night']



    @admin.display(description='Status')
    def status(self, obj):
        return ", ".join(list(obj.flag_labels))

    @admin.display(description='Total nº of raw files')
    def count_rawfits(self, obj):
        return obj.rawfits.count()
    
    @admin.display(description='Nº of bias')
    def count_bias(self, obj):
        return obj.rawbiascount
    
    @admin.display(description='Nº of flats')
    def count_flats(self, obj):
        return obj.rawflatcount
    
    @admin.display(description='Nº of darks')
    def count_darks(self, obj):
        return obj.rawdarkcount
    
    @admin.display(description='Nº of science files')
    def count_light(self, obj):
        return obj.rawlightcount
    
    @admin.display(description='Nº of reduced files')
    def count_reduced(self, obj):
        return obj.reducedcount
    
    @admin.display(description='Nº of calibrated files')
    def count_calibrated(self, obj):
        return obj.calibratedcount
        
    @admin.display(description='Summary of raw files')
    def summary_rawfits_status(self, obj):
        summary_rawfits_status = obj.get_summary_rawfits_status()
        return summary_rawfits_status
       
    @admin.display(description='Details')
    def details(self, obj):
        url = reverse('iop4admin:iop4api_epoch_details', args=[obj.id])
        url_rawfits = reverse(f'iop4admin:iop4api_rawfit_changelist') + f'?telescope={obj.telescope}&night={obj.night}'
        url_reducedfits = reverse(f'iop4admin:iop4api_reducedfit_changelist')+ f'?telescope={obj.telescope}&night={obj.night}'
        return format_html(rf'<a href="{url}">details</a> / '
                           rf'<a href="{url_rawfits}">rawfits</a> / '
                           rf'<a href="{url_reducedfits}">reducedfits</a>')
    

    def get_urls(self):
        urls = super().get_urls()
        my_urls = [
            path('details/<int:pk>', self.admin_site.admin_view(views.EpochDetailsView.as_view()), name='iop4api_epoch_details'),
        ]
        return my_urls + urls
    


# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

 # django imports
from django.db import models

# other imports
import os
import io
import numpy as np
import matplotlib as mplt
import matplotlib.pyplot as plt
from astropy.visualization.mpl_normalize import ImageNormalize
from astropy.visualization import LogStretch
from astropy.nddata import Cutout2D
from photutils.aperture import CircularAperture, CircularAnnulus

# iop4lib imports
from ..enums import *
from iop4lib.instruments.instrument import Instrument

# logging
import logging
logger = logging.getLogger(__name__)



class AperPhotResult(models.Model):
    """ Aperture photometry results for an AstroSource in a ReducedFit, either the Ordinary or Extraordinary pair."""

    # database fields

    ## identifiers

    reducedfit = models.ForeignKey("ReducedFit", on_delete=models.CASCADE, related_name='aperphotresults', help_text="The ReducedFit this AperPhotResult has been computed for.")
    astrosource = models.ForeignKey("AstroSource", on_delete=models.CASCADE, related_name='aperphotresults', help_text="The AstroSource this AperPhotResult has been computed for.")
    aperpix = models.FloatField(null=True, blank=True)
    r_in = models.FloatField(null=True, blank=True)
    r_out = models.FloatField(null=True, blank=True)
    pairs = models.TextField(null=True, blank=True, choices=PAIRS.choices, help_text="Whether this AperPhotResult is for the Ordinary or Extraordinary pair.")

    ## photometry results

    bkg_flux_counts = models.FloatField(null=True, blank=True)
    bkg_flux_counts_err = models.FloatField(null=True, blank=True)
    
    flux_counts = models.FloatField(null=True, blank=True)
    flux_counts_err = models.FloatField(null=True, blank=True)

    ## extra fields
    x_px = models.FloatField(null=True, help_text="used pixel position of the source in the image, x coordinate.")
    y_px = models.FloatField(null=True, help_text="used pixel position of the source in the image, y coordinate.")
    fwhm = models.FloatField(null=True, blank=True)
    photopolresults = models.ManyToManyField("PhotoPolResult", related_name='aperphotresults', help_text="The PhotoPolResult(s) this AperPhotResult has been used for.")
    modified = models.DateTimeField(auto_now=True)

    class Meta:
        app_label = 'iop4api'
        verbose_name = 'Aperture Photometry Result'
        verbose_name_plural = 'Aperture Photometry Results'
        constraints = [
            models.UniqueConstraint(fields=['reducedfit', 'astrosource', 'aperpix', 'r_in', 'r_out', 'pairs'], name='unique_aperphotresult')
        ]

    # repr and str

    def __repr__(self):
        return f'{self.__class__.__name__}.objects.get(id={self.id!r})'
    
    def __str__(self):
        return f'<{self.__class__.__name__} {self.id} | {self.reducedfit.fileloc} {self.astrosource.name} {self.aperpix} px {self.pairs}>' 

    @classmethod
    def create(cls, reducedfit, astrosource, aperpix, pairs, **kwargs):

        if (result := AperPhotResult.objects.filter(reducedfit=reducedfit, astrosource=astrosource, aperpix=aperpix, pairs=pairs).first()) is not None:
            logger.debug(f"AperPhotResult for {reducedfit}, {astrosource}, {aperpix}, {pairs} already exists, it will be used instead.")
        else:
            logger.debug(f"Creating AperPhotResult for {reducedfit}, {astrosource}, {aperpix}, {pairs}.")
            result = cls(reducedfit=reducedfit, astrosource=astrosource, aperpix=aperpix, pairs=pairs)

        for key, value in kwargs.items():
            setattr(result, key, value)

        result.save()

        return result

    @property
    def filedpropdir(self):
        return os.path.join(iop4conf.datadir, "aperphotresults", str(self.id))
    
    def get_img(self, force_rebuild=True, **kwargs):
        """
        Build an image preview (png) of the aperture and annulus over the source.

        If called with default arguments (no kwargs) it will try to load from the disk,
        except if called with force_rebuild.

        When called with default arguments (no kwargs), if rebuilt, it will save the image to disk.
        """
                     
        wcs = self.reducedfit.wcs1 if self.pairs == 'O' else self.reducedfit.wcs2

        if self.reducedfit.has_pairs:
            cutout_size = np.ceil(2.2*np.linalg.norm(Instrument.by_name(self.reducedfit.instrument).disp_sign_mean))
        else:
            cutout_size = np.ceil(1.3*self.r_out)

        cutout = Cutout2D(self.reducedfit.mdata, (self.x_px, self.y_px), (cutout_size, cutout_size), wcs)

        width = kwargs.get('width', 256)
        height = kwargs.get('height', 256)
        normtype = kwargs.get('norm', "log")
        vmin = kwargs.get('vmin', np.quantile(cutout.data.compressed(), 0.3))
        vmax = kwargs.get('vmax', np.quantile(cutout.data.compressed(), 0.99))
        a = kwargs.get('a', 10)

        fpath = os.path.join(self.filedpropdir, "img_preview_image.png")

        if len(kwargs) == 0 and not force_rebuild:
            if os.path.isfile(fpath) and os.path.getmtime(self.filepath) < os.path.getmtime(fpath):
                with open(fpath, 'rb') as f:
                    return f.read()

        cmap = plt.cm.gray.copy()
        cmap.set_bad(color='red')
        cmap.set_under(color='black')
        cmap.set_over(color='white')
            
        if normtype == "log":
            norm = ImageNormalize(cutout.data.compressed(), vmin=vmin, vmax=vmax, stretch=LogStretch(a=a))
        elif normtype == "logstretch":
            norm = ImageNormalize(stretch=LogStretch(a=a))


        buf = io.BytesIO()

        fig = mplt.figure.Figure(figsize=(width/100, height/100), dpi=iop4conf.mplt_default_dpi)
        ax = fig.subplots()

        wcs_px_pos = self.astrosource.coord.to_pixel(cutout.wcs)
        xy_px_pos = cutout.to_cutout_position((self.x_px, self.y_px))
        ap = CircularAperture(xy_px_pos, r=self.aperpix)
        annulus = CircularAnnulus(xy_px_pos, r_in=self.r_in, r_out=self.r_out)
    
        ax.imshow(cutout.data, cmap=cmap, origin='lower', norm=norm)
        ax.plot(wcs_px_pos[0], wcs_px_pos[1], 'rx', label='WCS')
        ax.plot(xy_px_pos[0], xy_px_pos[1], 'bo', label='Photometry')
        ap.plot(ax, color='blue', lw=2, alpha=1)
        annulus.plot(ax, color='green', lw=2, alpha=1)

        ax.axis('off')
        fig.savefig(buf, format='png', bbox_inches='tight', pad_inches=0)
        fig.clf()

        buf.seek(0)
        imgbytes = buf.read()

        # if it was rebuilt, save it to disk if it is the default image settings.
        if len(kwargs) == 0:
            if not os.path.exists(self.filedpropdir):
                os.makedirs(self.filedpropdir)
            with open(fpath, 'wb') as f:
                f.write(imgbytes)

        return imgbytes